@@ -21,9 +21,17 @@ enum wf_animation_type
 class animation_base
 {
   public:
-    virtual void init(wayfire_view view, int duration, wf_animation_type type);
+    virtual void init(wayfire_view view, wf::animation_description_t duration, wf_animation_type type);
     virtual bool step(); /* return true if continue, false otherwise */
+    virtual void reverse(); /* reverse the animation */
+    virtual int get_direction();
+
+    animation_base() = default;
     virtual ~animation_base();
+    animation_base(const animation_base &) = default;
+    animation_base(animation_base &&) = default;
+    animation_base& operator =(const animation_base&) = default;
+    animation_base& operator =(animation_base&&) = default;
 };
 
 #endif
