@@ -3,23 +3,23 @@
 
 #include <wayfire/view-transform.hpp>
 #include <wayfire/render-manager.hpp>
-#include <memory>
-
 #include "../animate.hpp"
 
 class FireTransformer;
+class ParticleSystem;
+
 class FireAnimation : public animation_base
 {
     std::string name; // the name of the transformer in the view's table
     wayfire_view view;
-    nonstd::observer_ptr<FireTransformer> transformer;
     wf::animation::simple_animation_t progression;
 
   public:
 
     ~FireAnimation();
-    void init(wayfire_view view, int duration, wf_animation_type type) override;
+    void init(wayfire_view view, wf::animation_description_t, wf_animation_type type) override;
     bool step() override; /* return true if continue, false otherwise */
+    void reverse() override; /* reverse the animation */
 };
 
 #endif /* end of include guard: FIRE_ANIMATION_HPP */
