@@ -1,12 +1,15 @@
+#include <wayfire/per-output-plugin.hpp>
+#include <memory>
 #include <wayfire/plugin.hpp>
 #include <wayfire/opengl.hpp>
 #include <wayfire/output.hpp>
 #include <wayfire/core.hpp>
 #include <wayfire/workspace-stream.hpp>
 #include <wayfire/render-manager.hpp>
-#include <wayfire/workspace-manager.hpp>
-
-#include <wayfire/plugins/common/workspace-stream-sharing.hpp>
+#include <wayfire/workspace-set.hpp>
+#include <wayfire/scene-operations.hpp>
+#include <wayfire/plugins/common/input-grab.hpp>
+#include "plugins/ipc/ipc-activator.hpp"
 
 #include <glm/gtc/matrix_transform.hpp>
 #include <wayfire/img.hpp>
@@ -16,6 +19,10 @@
 #include "skydome.hpp"
 #include "cubemap.hpp"
 #include "cube-control-signal.hpp"
+#include "wayfire/region.hpp"
+#include "wayfire/scene-render.hpp"
+#include "wayfire/scene.hpp"
+#include "wayfire/signal-definitions.hpp"
 
 #define Z_OFFSET_NEAR 0.89567f
 #define Z_OFFSET_FAR  2.00000f
@@ -30,13 +37,157 @@
 #include "shaders.tpp"
 #include "shaders-3-2.tpp"
 
-class wayfire_cube : public wf::plugin_interface_t
+class wayfire_cube : public wf::per_output_plugin_instance_t, public wf::pointer_interaction_t
 {
-    wf::button_callback activate_binding;
-    wf::activator_callback rotate_left, rotate_right;
-    wf::render_hook_t renderer;
+    class cube_render_node_t : public wf::scene::node_t
+    {
+        class cube_render_instance_t : public wf::scene::render_instance_t
+        {
+            std::shared_ptr<cube_render_node_t> self;
+            wf::scene::damage_callback push_damage;
 
-    nonstd::observer_ptr<wf::workspace_stream_pool_t> streams;
+            std::vector<std::vector<wf::scene::render_instance_uptr>> ws_instances;
+            std::vector<wf::region_t> ws_damage;
+            std::vector<wf::render_target_t> framebuffers;
+
+            wf::signal::connection_t<wf::scene::node_damage_signal> on_cube_damage =
+                [=] (wf::scene::node_damage_signal *ev)
+            {
+                push_damage(ev->region);
+            };
+
+          public:
+            cube_render_instance_t(cube_render_node_t *self, wf::scene::damage_callback push_damage)
+            {
+                this->self = std::dynamic_pointer_cast<cube_render_node_t>(self->shared_from_this());
+                this->push_damage = push_damage;
+                self->connect(&on_cube_damage);
+
+                ws_damage.resize(self->workspaces.size());
+                framebuffers.resize(self->workspaces.size());
+                ws_instances.resize(self->workspaces.size());
+                for (int i = 0; i < (int)self->workspaces.size(); i++)
+                {
+                    auto push_damage_child = [=] (const wf::region_t& damage)
+                    {
+                        ws_damage[i] |= damage;
+                        push_damage(self->get_bounding_box());
+                    };
+
+                    self->workspaces[i]->gen_render_instances(ws_instances[i],
+                        push_damage_child, self->cube->output);
+
+                    ws_damage[i] |= self->workspaces[i]->get_bounding_box();
+                }
+            }
+
+            ~cube_render_instance_t()
+            {
+                OpenGL::render_begin();
+                for (auto& buf : framebuffers)
+                {
+                    buf.release();
+                }
+
+                OpenGL::render_end();
+            }
+
+            void schedule_instructions(
+                std::vector<wf::scene::render_instruction_t>& instructions,
+                const wf::render_target_t& target, wf::region_t& damage) override
+            {
+                instructions.push_back(wf::scene::render_instruction_t{
+                    .instance = this,
+                    .target   = target,
+                    .damage   = damage & self->get_bounding_box(),
+                });
+
+                auto bbox = self->get_bounding_box();
+
+                damage ^= bbox;
+            }
+
+            void render(const wf::render_target_t& target,
+                const wf::region_t& region, const std::any& tag) override
+            {
+                for (int i = 0; i < (int)ws_instances.size(); i++)
+                {
+                    framebuffers[i].geometry = self->workspaces[i]->get_bounding_box();
+                    framebuffers[i].scale    = self->cube->output->handle->scale;
+                    framebuffers[i].wl_transform = WL_OUTPUT_TRANSFORM_NORMAL;
+                    framebuffers[i].transform    = get_output_matrix_from_transform(
+                        framebuffers[i].wl_transform);
+
+                    auto size = framebuffers[i].framebuffer_box_from_geometry_box(framebuffers[i].geometry);
+                    OpenGL::render_begin();
+                    framebuffers[i].allocate(size.width, size.height);
+                    OpenGL::render_end();
+
+                    wf::scene::render_pass_params_t params;
+                    params.instances = &ws_instances[i];
+                    params.damage    = ws_damage[i];
+                    params.reference_output = self->cube->output;
+                    params.target = framebuffers[i];
+                    wf::scene::run_render_pass(params, wf::scene::RPASS_CLEAR_BACKGROUND |
+                        wf::scene::RPASS_EMIT_SIGNALS);
+
+                    ws_damage[i].clear();
+                }
+
+                self->cube->render(target.translated(-wf::origin(self->get_bounding_box())), framebuffers);
+            }
+
+            void compute_visibility(wf::output_t *output, wf::region_t& visible) override
+            {
+                for (int i = 0; i < (int)self->workspaces.size(); i++)
+                {
+                    wf::region_t ws_region = self->workspaces[i]->get_bounding_box();
+                    for (auto& ch : this->ws_instances[i])
+                    {
+                        ch->compute_visibility(output, ws_region);
+                    }
+                }
+            }
+        };
+
+      public:
+        cube_render_node_t(wayfire_cube *cube) : node_t(false)
+        {
+            this->cube = cube;
+            auto w = cube->output->wset()->get_workspace_grid_size().width;
+            auto y = cube->output->wset()->get_current_workspace().y;
+            for (int i = 0; i < w; i++)
+            {
+                auto node = std::make_shared<wf::workspace_stream_node_t>(cube->output, wf::point_t{i, y});
+                workspaces.push_back(node);
+            }
+        }
+
+        virtual void gen_render_instances(
+            std::vector<wf::scene::render_instance_uptr>& instances,
+            wf::scene::damage_callback push_damage, wf::output_t *shown_on)
+        {
+            if (shown_on != this->cube->output)
+            {
+                return;
+            }
+
+            instances.push_back(std::make_unique<cube_render_instance_t>(
+                this, push_damage));
+        }
+
+        wf::geometry_t get_bounding_box()
+        {
+            return cube->output->get_layout_geometry();
+        }
+
+      private:
+        std::vector<std::shared_ptr<wf::workspace_stream_node_t>> workspaces;
+        wayfire_cube *cube;
+    };
+
+    std::unique_ptr<wf::input_grab_t> input_grab;
+    std::shared_ptr<cube_render_node_t> render_node;
 
     wf::option_wrapper_t<double> XVelocity{"cube/speed_spin_horiz"},
     YVelocity{"cube/speed_spin_vert"}, ZVelocity{"cube/speed_zoom"};
@@ -52,10 +203,6 @@ class wayfire_cube : public wf::plugin_interface_t
     wf::option_wrapper_t<bool> use_light{"cube/light"};
     wf::option_wrapper_t<int> use_deform{"cube/deform"};
 
-    wf::option_wrapper_t<wf::buttonbinding_t> button{"cube/activate"};
-    wf::option_wrapper_t<wf::activatorbinding_t> key_left{"cube/rotate_left"};
-    wf::option_wrapper_t<wf::activatorbinding_t> key_right{"cube/rotate_right"};
-
     std::string last_background_mode;
     std::unique_ptr<wf_cube_background_base> background;
 
@@ -63,7 +210,7 @@ class wayfire_cube : public wf::plugin_interface_t
 
     void reload_background()
     {
-        if (!last_background_mode.compare(background_mode))
+        if (last_background_mode == (std::string)background_mode)
         {
             return;
         }
@@ -91,14 +238,20 @@ class wayfire_cube : public wf::plugin_interface_t
 
     int get_num_faces()
     {
-        return output->workspace->get_workspace_grid_size().width;
+        return output->wset()->get_workspace_grid_size().width;
     }
 
+    wf::plugin_activation_data_t grab_interface{
+        .name = "cube",
+        .capabilities = wf::CAPABILITY_MANAGE_COMPOSITOR,
+        .cancel = [=] () { deactivate(); },
+    };
+
   public:
     void init() override
     {
-        grab_interface->name = "cube";
-        grab_interface->capabilities = wf::CAPABILITY_MANAGE_COMPOSITOR;
+        input_grab = std::make_unique<wf::input_grab_t>("cube", output, nullptr, this, nullptr);
+        input_grab->set_wants_raw_input(true);
 
         animation.cube_animation.offset_y.set(0, 0);
         animation.cube_animation.offset_z.set(0, 0);
@@ -110,74 +263,34 @@ class wayfire_cube : public wf::plugin_interface_t
 
         reload_background();
 
-        activate_binding = [=] (auto)
-        {
-            return input_grabbed();
-        };
+        output->connect(&on_cube_control);
 
-        rotate_left = [=] (auto)
-        {
-            return move_vp(-1);
-        };
-
-        rotate_right = [=] (auto)
-        {
-            return move_vp(1);
-        };
-
-        output->add_button(button, &activate_binding);
-        output->add_activator(key_left, &rotate_left);
-        output->add_activator(key_right, &rotate_right);
-        output->connect_signal("cube-control", &on_cube_control);
-
-        grab_interface->callbacks.pointer.button = [=] (uint32_t b, uint32_t s)
-        {
-            if (s == WL_POINTER_BUTTON_STATE_RELEASED)
-            {
-                input_ungrabbed();
-            }
-        };
-
-        grab_interface->callbacks.pointer.axis = [=] (
-            wlr_pointer_axis_event *ev)
-        {
-            if (ev->orientation == WLR_AXIS_ORIENTATION_VERTICAL)
-            {
-                pointer_scrolled(ev->delta);
-            }
-        };
-
-        grab_interface->callbacks.cancel = [=] ()
-        {
-            deactivate();
-        };
-
-        auto wsize = output->workspace->get_workspace_grid_size();
-        animation.side_angle = 2 * M_PI / float(wsize.width);
-        identity_z_offset    = 0.5 / std::tan(animation.side_angle / 2);
-        if (wsize.width == 1)
-        {
-            // tan(M_PI) is 0, so identity_z_offset is invalid
-            identity_z_offset = 0.0f;
-        }
-
-        animation.cube_animation.offset_z.set(identity_z_offset + Z_OFFSET_NEAR,
-            identity_z_offset + Z_OFFSET_NEAR);
-
-        renderer = [=] (const wf::framebuffer_t& dest) {render(dest);};
-
-        OpenGL::render_begin(output->render->get_target_framebuffer());
+        OpenGL::render_begin();
         load_program();
         OpenGL::render_end();
     }
 
+    void handle_pointer_button(const wlr_pointer_button_event& event) override
+    {
+        if (event.state == WLR_BUTTON_RELEASED)
+        {
+            input_ungrabbed();
+        }
+    }
+
+    void handle_pointer_axis(const wlr_pointer_axis_event& event) override
+    {
+        if (event.orientation == WLR_AXIS_ORIENTATION_VERTICAL)
+        {
+            pointer_scrolled(event.delta);
+        }
+    }
+
     void load_program()
     {
 #ifdef USE_GLES32
-        std::string ext_string(reinterpret_cast<const char*>(glGetString(
-            GL_EXTENSIONS)));
-        tessellation_support =
-            ext_string.find(std::string("GL_EXT_tessellation_shader")) !=
+        std::string ext_string(reinterpret_cast<const char*>(glGetString(GL_EXTENSIONS)));
+        tessellation_support = ext_string.find(std::string("GL_EXT_tessellation_shader")) !=
             std::string::npos;
 #else
         tessellation_support = false;
@@ -185,8 +298,7 @@ class wayfire_cube : public wf::plugin_interface_t
 
         if (!tessellation_support)
         {
-            program.set_simple(OpenGL::compile_program(
-                cube_vertex_2_0, cube_fragment_2_0));
+            program.set_simple(OpenGL::compile_program(cube_vertex_2_0, cube_fragment_2_0));
         } else
         {
 #ifdef USE_GLES32
@@ -217,13 +329,11 @@ class wayfire_cube : public wf::plugin_interface_t
 #endif
         }
 
-        streams = wf::workspace_stream_pool_t::ensure_pool(output);
         animation.projection = glm::perspective(45.0f, 1.f, 0.1f, 100.f);
     }
 
-    wf::signal_callback_t on_cube_control = [=] (wf::signal_data_t *data)
+    wf::signal::connection_t<cube_control_signal> on_cube_control = [=] (cube_control_signal *d)
     {
-        cube_control_signal *d = dynamic_cast<cube_control_signal*>(data);
         rotate_and_zoom_cube(d->angle, d->zoom, d->ease, d->last_frame);
         d->carried_out = true;
     };
@@ -260,22 +370,38 @@ class wayfire_cube : public wf::plugin_interface_t
     /* Tries to initialize renderer, activate plugin, etc. */
     bool activate()
     {
-        if (output->is_plugin_active(grab_interface->name))
+        if (output->is_plugin_active(grab_interface.name))
         {
             return true;
         }
 
-        if (!output->activate_plugin(grab_interface))
+        if (!output->activate_plugin(&grab_interface))
         {
             return false;
         }
 
-        wf::get_core().connect_signal("pointer_motion", &on_motion_event);
-        output->render->set_renderer(renderer);
-        output->render->schedule_redraw();
-        wf::get_core().hide_cursor();
-        grab_interface->grab();
+        wf::get_core().connect(&on_motion_event);
 
+        render_node = std::make_shared<cube_render_node_t>(this);
+        wf::scene::add_front(wf::get_core().scene(), render_node);
+        output->render->add_effect(&pre_hook, wf::OUTPUT_EFFECT_PRE);
+        output->render->set_require_depth_buffer(true);
+
+        wf::get_core().hide_cursor();
+        input_grab->grab_input(wf::scene::layer::OVERLAY);
+
+        auto wsize = output->wset()->get_workspace_grid_size();
+        animation.side_angle = 2 * M_PI / float(wsize.width);
+        identity_z_offset    = 0.5 / std::tan(animation.side_angle / 2);
+        if (wsize.width == 1)
+        {
+            // tan(M_PI) is 0, so identity_z_offset is invalid
+            identity_z_offset = 0.0f;
+        }
+
+        reload_background();
+        animation.cube_animation.offset_z.set(identity_z_offset + Z_OFFSET_NEAR,
+            identity_z_offset + Z_OFFSET_NEAR);
         return true;
     }
 
@@ -289,34 +415,32 @@ class wayfire_cube : public wf::plugin_interface_t
     /* Disable custom rendering and deactivate plugin */
     void deactivate()
     {
-        if (!output->is_plugin_active(grab_interface->name))
+        if (!output->is_plugin_active(grab_interface.name))
         {
             return;
         }
 
-        output->render->set_renderer(nullptr);
+        wf::scene::remove_child(render_node);
+        render_node = nullptr;
+        output->render->rem_effect(&pre_hook);
+        output->render->set_require_depth_buffer(false);
 
-        grab_interface->ungrab();
-        output->deactivate_plugin(grab_interface);
+        input_grab->ungrab_input();
+        output->deactivate_plugin(&grab_interface);
         wf::get_core().unhide_cursor();
-        wf::get_core().disconnect_signal("pointer_motion", &on_motion_event);
+        on_motion_event.disconnect();
 
         /* Figure out how much we have rotated and switch workspace */
         int size = get_num_faces();
         int dvx  = calculate_viewport_dx_from_rotation();
 
-        auto cws = output->workspace->get_current_workspace();
+        auto cws = output->wset()->get_current_workspace();
         int nvx  = (cws.x + (dvx % size) + size) % size;
-        output->workspace->set_workspace({nvx, cws.y});
+        output->wset()->set_workspace({nvx, cws.y});
 
         /* We are finished with rotation, make sure the next time cube is used
          * it is properly reset */
         animation.cube_animation.rotation.set(0, 0);
-
-        for (int i = 0; i < size; i++)
-        {
-            streams->stop({i, cws.y});
-        }
     }
 
     /* Sets attributes target to such values that the cube effect isn't visible,
@@ -389,7 +513,8 @@ class wayfire_cube : public wf::plugin_interface_t
         update_view_matrix();
         output->render->schedule_redraw();
 
-        return true;
+        // Let the button go to the input grab
+        return false;
     }
 
     /* Mouse grab was released */
@@ -429,16 +554,7 @@ class wayfire_cube : public wf::plugin_interface_t
         animation.view = zoom_translate * rotation * view;
     }
 
-    void update_workspace_streams()
-    {
-        auto cws = output->workspace->get_current_workspace();
-        for (int i = 0; i < get_num_faces(); i++)
-        {
-            streams->update({i, cws.y});
-        }
-    }
-
-    glm::mat4 calculate_vp_matrix(const wf::framebuffer_t& dest)
+    glm::mat4 calculate_vp_matrix(const wf::render_target_t& dest)
     {
         float zoom_factor = animation.cube_animation.zoom;
         auto scale_matrix = glm::scale(glm::mat4(1.0),
@@ -448,7 +564,7 @@ class wayfire_cube : public wf::plugin_interface_t
     }
 
     /* Calculate the base model matrix for the i-th side of the cube */
-    glm::mat4 calculate_model_matrix(int i, glm::mat4 fb_transform)
+    glm::mat4 calculate_model_matrix(int i)
     {
         const float angle =
             i * animation.side_angle + animation.cube_animation.rotation;
@@ -467,23 +583,23 @@ class wayfire_cube : public wf::plugin_interface_t
         auto translation = glm::translate(glm::mat4(1.0),
             glm::vec3(0, 0, identity_z_offset + additional_z));
 
-        return rotation * translation * glm::inverse(fb_transform);
+        return rotation * translation;
     }
 
     /* Render the sides of the cube, using the given culling mode - cw or ccw */
-    void render_cube(GLuint front_face, glm::mat4 fb_transform)
+    void render_cube(GLuint front_face, glm::mat4 fb_transform,
+        const std::vector<wf::render_target_t>& buffers)
     {
         GL_CALL(glFrontFace(front_face));
         static const GLuint indexData[] = {0, 1, 2, 0, 2, 3};
 
-        auto cws = output->workspace->get_current_workspace();
+        auto cws = output->wset()->get_current_workspace();
         for (int i = 0; i < get_num_faces(); i++)
         {
             int index = (cws.x + i) % get_num_faces();
-            GL_CALL(glBindTexture(GL_TEXTURE_2D,
-                streams->get({index, cws.y}).buffer.tex));
+            GL_CALL(glBindTexture(GL_TEXTURE_2D, buffers[index].tex));
 
-            auto model = calculate_model_matrix(i, fb_transform);
+            auto model = calculate_model_matrix(i);
             program.uniformMatrix4f("model", model);
 
             if (tessellation_support)
@@ -499,9 +615,8 @@ class wayfire_cube : public wf::plugin_interface_t
         }
     }
 
-    void render(const wf::framebuffer_t& dest)
+    void render(const wf::render_target_t& dest, const std::vector<wf::render_target_t>& buffers)
     {
-        update_workspace_streams();
         if (program.get_program_id(wf::TEXTURE_TYPE_RGBA) == 0)
         {
             load_program();
@@ -510,8 +625,6 @@ class wayfire_cube : public wf::plugin_interface_t
         OpenGL::render_begin(dest);
         GL_CALL(glClear(GL_DEPTH_BUFFER_BIT));
         OpenGL::render_end();
-
-        reload_background();
         background->render_frame(dest, animation);
 
         auto vp = calculate_vp_matrix(dest);
@@ -551,16 +664,19 @@ class wayfire_cube : public wf::plugin_interface_t
          * that are on the back, and then we render those at the front, so we
          * don't have to use depth testing and we also can support alpha cube. */
         GL_CALL(glEnable(GL_CULL_FACE));
-        render_cube(GL_CCW, dest.transform);
-        render_cube(GL_CW, dest.transform);
+        render_cube(GL_CCW, dest.transform, buffers);
+        render_cube(GL_CW, dest.transform, buffers);
         GL_CALL(glDisable(GL_CULL_FACE));
 
         GL_CALL(glDisable(GL_DEPTH_TEST));
         program.deactivate();
         OpenGL::render_end();
+    }
 
+    wf::effect_hook_t pre_hook = [=] ()
+    {
         update_view_matrix();
-
+        wf::scene::damage_node(render_node, render_node->get_bounding_box());
         if (animation.cube_animation.running())
         {
             output->render->schedule_redraw();
@@ -568,13 +684,11 @@ class wayfire_cube : public wf::plugin_interface_t
         {
             deactivate();
         }
-    }
+    };
 
-    wf::signal_callback_t on_motion_event = [=] (wf::signal_data_t *data)
+    wf::signal::connection_t<wf::input_event_signal<wlr_pointer_motion_event>> on_motion_event =
+        [=] (wf::input_event_signal<wlr_pointer_motion_event> *ev)
     {
-        auto ev = static_cast<
-            wf::input_event_signal<wlr_pointer_motion_event>*>(data);
-
         pointer_moved(ev->event);
 
         ev->event->delta_x    = 0;
@@ -645,22 +759,53 @@ class wayfire_cube : public wf::plugin_interface_t
 
     void fini() override
     {
-        if (output->is_plugin_active(grab_interface->name))
+        if (output->is_plugin_active(grab_interface.name))
         {
             deactivate();
         }
 
-        streams->unref();
-
         OpenGL::render_begin();
         program.free_resources();
         OpenGL::render_end();
-
-        output->rem_binding(&activate_binding);
-        output->rem_binding(&rotate_left);
-        output->rem_binding(&rotate_right);
-        output->disconnect_signal("cube-control", &on_cube_control);
     }
 };
 
-DECLARE_WAYFIRE_PLUGIN(wayfire_cube);
+
+class wayfire_cube_global : public wf::plugin_interface_t,
+    public wf::per_output_tracker_mixin_t<wayfire_cube>
+{
+    wf::ipc_activator_t rotate_left{"cube/rotate_left"};
+    wf::ipc_activator_t rotate_right{"cube/rotate_right"};
+    wf::ipc_activator_t activate{"cube/activate"};
+
+  public:
+    void init() override
+    {
+        this->init_output_tracking();
+        rotate_left.set_handler(rotate_left_cb);
+        rotate_right.set_handler(rotate_right_cb);
+        activate.set_handler(activate_cb);
+    }
+
+    void fini() override
+    {
+        this->fini_output_tracking();
+    }
+
+    wf::ipc_activator_t::handler_t rotate_left_cb = [=] (wf::output_t *output, wayfire_view)
+    {
+        return this->output_instance[output]->move_vp(-1);
+    };
+
+    wf::ipc_activator_t::handler_t rotate_right_cb = [=] (wf::output_t *output, wayfire_view)
+    {
+        return this->output_instance[output]->move_vp(+1);
+    };
+
+    wf::ipc_activator_t::handler_t activate_cb = [=] (wf::output_t *output, wayfire_view)
+    {
+        return this->output_instance[output]->input_grabbed();
+    };
+};
+
+DECLARE_WAYFIRE_PLUGIN(wayfire_cube_global);
