@@ -141,13 +141,16 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
 
     std::optional<wf::scene::input_node_t> find_node_at(const wf::pointf_t& at) override
     {
-        wf::pointf_t local = at - wf::pointf_t{get_offset()};
-        if (cached_region.contains_pointf(local))
+        if (auto view = _view.lock())
         {
-            return wf::scene::input_node_t{
-                .node = this,
-                .local_coords = local,
-            };
+            wf::pointf_t local = at - wf::pointf_t{get_offset()};
+            if (cached_region.contains_pointf(local) && view->is_mapped())
+            {
+                return wf::scene::input_node_t{
+                    .node = this,
+                    .local_coords = local,
+                };
+            }
         }
 
         return {};
@@ -165,7 +168,7 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
 
     class decoration_render_instance_t : public wf::scene::render_instance_t
     {
-        simple_decoration_node_t *self;
+        std::shared_ptr<simple_decoration_node_t> self;
         wf::scene::damage_callback push_damage;
 
         wf::signal::connection_t<wf::scene::node_damage_signal> on_surface_damage =
@@ -177,7 +180,7 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
       public:
         decoration_render_instance_t(simple_decoration_node_t *self, wf::scene::damage_callback push_damage)
         {
-            this->self = self;
+            this->self = std::dynamic_pointer_cast<simple_decoration_node_t>(self->shared_from_this());
             this->push_damage = push_damage;
             self->connect(&on_surface_damage);
         }
