@@ -1,13 +1,16 @@
-#include <cctype>
+#include "wayfire/plugins/common/shared-core-data.hpp"
+#include "wayfire/util.hpp"
 #include <string>
 #include <map>
 #include <wayfire/plugin.hpp>
-#include <wayfire/singleton-plugin.hpp>
+#include <wayfire/per-output-plugin.hpp>
 #include <wayfire/output.hpp>
 #include <wayfire/signal-definitions.hpp>
 #include <wayfire/util/log.hpp>
 #include <wayfire/plugins/scale-signal.hpp>
 #include <wayfire/nonstd/wlroots-full.hpp>
+#include <wayfire/toplevel-view.hpp>
+#include <wayfire/seat.hpp>
 
 #include <linux/input-event-codes.h>
 
@@ -15,31 +18,7 @@
 #include <wayfire/opengl.hpp>
 #include <wayfire/plugins/common/cairo-util.hpp>
 #include <wayfire/plugins/common/simple-texture.hpp>
-
-struct scale_key_repeat_t
-{
-    wf::option_wrapper_t<int> delay{"input/kb_repeat_delay"};
-    wf::option_wrapper_t<int> rate{"input/kb_repeat_rate"};
-
-    wf::wl_timer timer_delay;
-    wf::wl_timer timer_rate;
-
-    using callback_t = std::function<void (uint32_t)>;
-
-    scale_key_repeat_t(uint32_t key, callback_t handler)
-    {
-        timer_delay.set_timeout(delay, [=] ()
-        {
-            timer_rate.set_timeout(1000 / rate, [=] ()
-            {
-                handler(key);
-                return true; // repeat
-            });
-
-            return false; // no more repeat
-        });
-    }
-};
+#include <wayfire/plugins/common/key-repeat.hpp>
 
 class scale_title_filter;
 
@@ -97,11 +76,12 @@ struct scale_title_filter_text
     }
 };
 
-class scale_title_filter : public wf::singleton_plugin_t<scale_title_filter_text>
+class scale_title_filter : public wf::per_output_plugin_instance_t
 {
     wf::option_wrapper_t<bool> case_sensitive{"scale-title-filter/case_sensitive"};
     wf::option_wrapper_t<bool> share_filter{"scale-title-filter/share_filter"};
     scale_title_filter_text local_filter;
+    wf::shared_data::ref_ptr_t<scale_title_filter_text> global_filter;
 
     inline void fix_case(std::string& string)
     {
@@ -144,7 +124,7 @@ class scale_title_filter : public wf::singleton_plugin_t<scale_title_filter_text
 
     scale_title_filter_text& get_active_filter()
     {
-        return share_filter ? get_instance() : local_filter;
+        return share_filter ? *global_filter.get() : local_filter;
     }
 
   public:
@@ -157,54 +137,41 @@ class scale_title_filter : public wf::singleton_plugin_t<scale_title_filter_text
 
     void init() override
     {
-        wf::singleton_plugin_t<scale_title_filter_text>::init();
-
-        auto& global = get_instance();
-        global.add_instance(this);
-
-        grab_interface->name = "scale-title-filter";
-        grab_interface->capabilities = 0;
-
+        global_filter->add_instance(this);
         share_filter.set_callback(shared_option_changed);
-        output->connect_signal("scale-filter", &view_filter);
-        output->connect_signal("scale-end", &scale_end);
+        output->connect(&view_filter);
+        output->connect(&scale_end);
     }
 
     void fini() override
     {
         do_end_scale();
-
-        auto& global = get_instance();
-        global.rem_instance(this);
-
-        wf::singleton_plugin_t<scale_title_filter_text>::fini();
+        global_filter->rem_instance(this);
     }
 
-    wf::signal_connection_t view_filter{[this] (wf::signal_data_t *data)
+    wf::signal::connection_t<scale_filter_signal> view_filter = [=] (scale_filter_signal *ev)
+    {
+        if (!scale_running)
         {
-            if (!scale_running)
-            {
-                wf::get_core().connect_signal("keyboard_key", &scale_key);
-                scale_running = true;
-                update_overlay();
-            }
-
-            auto signal = static_cast<scale_filter_signal*>(data);
-            scale_filter_views(signal, [this] (wayfire_view v)
-            {
-                return !should_show_view(v);
-            });
+            wf::get_core().connect(&scale_key);
+            scale_running = true;
+            update_overlay();
         }
+
+        scale_filter_views(ev, [this] (wayfire_toplevel_view v)
+        {
+            return !should_show_view(v);
+        });
     };
 
-    std::map<uint32_t, std::unique_ptr<scale_key_repeat_t>> keys;
-    scale_key_repeat_t::callback_t handle_key_repeat = [=] (uint32_t raw_keycode)
+    std::map<uint32_t, std::unique_ptr<wf::key_repeat_t>> keys;
+    wf::key_repeat_t::callback_t handle_key_repeat = [=] (uint32_t raw_keycode)
     {
         auto seat     = wf::get_core().get_current_seat();
         auto keyboard = wlr_seat_get_keyboard(seat);
         if (!keyboard)
         {
-            return; /* should not happen */
+            return false; /* should not happen */
         }
 
         auto xkb_state = keyboard->xkb_state;
@@ -218,52 +185,58 @@ class scale_title_filter : public wf::singleton_plugin_t<scale_title_filter_text
         {
             filter.add_key(xkb_state, keycode);
         }
+
+        return true;
     };
 
+    wf::wl_idle_call idle_update_filter;
+
     void update_filter()
     {
-        if (scale_running)
+        // Delay updating the filter in case the last key causes scale to exit.
+        idle_update_filter.run_once([&]
         {
-            output->emit_signal("scale-update", nullptr);
-            update_overlay();
-        }
+            if (scale_running)
+            {
+                scale_update_signal ev;
+                output->emit(&ev);
+                update_overlay();
+            }
+        });
     }
 
-    wf::signal_connection_t scale_key = [this] (wf::signal_data_t *data)
+    wf::signal::connection_t<wf::input_event_signal<wlr_keyboard_key_event>> scale_key =
+        [this] (wf::input_event_signal<wlr_keyboard_key_event> *ev)
     {
-        auto k =
-            static_cast<wf::input_event_signal<wlr_keyboard_key_event>*>(data);
-        if (k->event->state == WL_KEYBOARD_KEY_STATE_RELEASED)
+        if (ev->event->state == WL_KEYBOARD_KEY_STATE_RELEASED)
         {
-            keys.erase(k->event->keycode);
+            keys.erase(ev->event->keycode);
             return;
         }
 
-        if ((k->event->keycode == KEY_ESC) || (k->event->keycode == KEY_ENTER))
+        if ((ev->event->keycode == KEY_ESC) || (ev->event->keycode == KEY_ENTER))
         {
             return;
         }
 
-        if (output != wf::get_core().get_active_output())
+        if (output != wf::get_core().seat->get_active_output())
         {
             return;
         }
 
-        keys[k->event->keycode] =
-            std::make_unique<scale_key_repeat_t>(k->event->keycode,
-                handle_key_repeat);
-        handle_key_repeat(k->event->keycode);
+        keys[ev->event->keycode] = std::make_unique<wf::key_repeat_t>(ev->event->keycode, handle_key_repeat);
+        handle_key_repeat(ev->event->keycode);
     };
 
 
-    wf::signal_connection_t scale_end = [this] (wf::signal_data_t*)
+    wf::signal::connection_t<scale_end_signal> scale_end = [=] (scale_end_signal *ev)
     {
         do_end_scale();
     };
 
     void do_end_scale()
     {
-        wf::get_core().disconnect_signal(&scale_key);
+        scale_key.disconnect();
         keys.clear();
         clear_overlay();
         scale_running = false;
@@ -275,9 +248,11 @@ class scale_title_filter : public wf::singleton_plugin_t<scale_title_filter_text
         if (scale_running)
         {
             /* clear the filter that is not used anymore */
-            auto& filter = share_filter ? local_filter : get_instance();
+            auto& filter = get_active_filter();
             filter.clear();
-            output->emit_signal("scale-update", nullptr);
+
+            scale_update_signal ev;
+            output->emit(&ev);
             update_overlay();
         }
     };
@@ -469,4 +444,4 @@ void scale_title_filter_text::check_scale_end()
     }
 }
 
-DECLARE_WAYFIRE_PLUGIN(scale_title_filter);
+DECLARE_WAYFIRE_PLUGIN(wf::per_output_plugin_t<scale_title_filter>);
