@@ -50,7 +50,7 @@ class scale_animation_t : public duration_t
 
 struct wf_scale_animation_attribs
 {
-    wf::option_wrapper_t<int> duration{"scale/duration"};
+    wf::option_wrapper_t<wf::animation_description_t> duration{"scale/duration"};
     scale_animation_t scale_animation{duration};
 };
 
@@ -100,10 +100,10 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
     wf::point_t initial_workspace;
     bool hook_set;
     /* View that was active before scale began. */
-    wayfire_toplevel_view initial_focus_view;
+    std::weak_ptr<wf::view_interface_t> initial_focus_view;
     /* View that has active focus. */
     wayfire_toplevel_view current_focus_view;
-    // View over which the last input press happened, might become dangling
+    // View over which the last input press happened
     wayfire_toplevel_view last_selected_view;
     std::map<wayfire_toplevel_view, view_scale_data> scale_data;
     wf::option_wrapper_t<int> spacing{"scale/spacing"};
@@ -113,6 +113,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
     wf::option_wrapper_t<double> minimized_alpha{"scale/minimized_alpha"};
     wf::option_wrapper_t<bool> allow_scale_zoom{"scale/allow_zoom"};
     wf::option_wrapper_t<bool> include_minimized{"scale/include_minimized"};
+    wf::option_wrapper_t<bool> close_on_new_view{"scale/close_on_new_view"};
 
     /* maximum scale -- 1.0 means we will not "zoom in" on a view */
     const double max_scale_factor = 1.0;
@@ -181,7 +182,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
 
             // vswitch picks the top view, we want the focused one
             std::vector<wayfire_toplevel_view> fixed_views;
-            if (view && !all_workspaces)
+            if (view && current_focus_view && !all_workspaces)
             {
                 fixed_views.push_back(current_focus_view);
             }
@@ -216,6 +217,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         /* Transformers are added only once when scale is activated so
          * this is a good place to connect the geometry-changed handler */
         view->connect(&view_geometry_changed);
+        view->connect(&view_unmapped);
 
         set_tiled_wobbly(view, true);
 
@@ -235,6 +237,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         data.view = view;
         output->emit(&data);
         view->get_transformed_node()->rem_transformer("scale");
+        view->disconnect(&view_unmapped);
         set_tiled_wobbly(view, false);
     }
 
@@ -332,17 +335,6 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         }
     }
 
-    /** Return the topmost parent */
-    wayfire_toplevel_view get_top_parent(wayfire_toplevel_view view)
-    {
-        while (view && view->parent)
-        {
-            view = view->parent;
-        }
-
-        return view;
-    }
-
     /* Fade all views' alpha to inactive alpha except the
      * view argument */
     void fade_out_all_except(wayfire_toplevel_view view)
@@ -350,7 +342,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         for (auto& e : scale_data)
         {
             auto v = e.first;
-            if (get_top_parent(v) == get_top_parent(view))
+            if (wf::find_topmost_parent(v) == wf::find_topmost_parent(view))
             {
                 continue;
             }
@@ -421,19 +413,19 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
     {
         if (view == current_focus_view)
         {
-            current_focus_view = toplevel_cast(wf::get_active_view_for_output(output));
+            current_focus_view = nullptr;
         }
 
-        if (view == initial_focus_view)
+        if (view == last_selected_view)
         {
-            initial_focus_view = nullptr;
+            last_selected_view = nullptr;
         }
     }
 
     /* Remove transformer from view and remove view from the scale_data map */
     void remove_view(wayfire_toplevel_view view)
     {
-        if (!view)
+        if (!view || !scale_data.count(view))
         {
             return;
         }
@@ -452,8 +444,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
     }
 
     /* Process button event */
-    void process_input(uint32_t button, uint32_t state,
-        wf::pointf_t input_position)
+    void process_input(uint32_t button, uint32_t state, wf::pointf_t input_position)
     {
         if (!active)
         {
@@ -493,10 +484,10 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
             // Focus the view under the mouse
             current_focus_view = view;
             fade_out_all_except(view);
-            fade_in(get_top_parent(view));
+            fade_in(wf::find_topmost_parent(view));
 
             // End scale
-            initial_focus_view = nullptr;
+            initial_focus_view.reset();
             deactivate();
             break;
 
@@ -546,11 +537,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
     /* Get the workspace for the center point of the untransformed view geometry */
     wf::point_t get_view_main_workspace(wayfire_toplevel_view view)
     {
-        while (view->parent)
-        {
-            view = view->parent;
-        }
-
+        view = wf::find_topmost_parent(view);
         auto ws     = output->wset()->get_current_workspace();
         auto og     = output->get_layout_geometry();
         auto vg     = view->get_geometry();
@@ -641,9 +628,13 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
           case KEY_ESC:
             deactivate();
             output->wset()->request_workspace(initial_workspace);
-            wf::get_core().default_wm->focus_raise_view(initial_focus_view);
-            initial_focus_view = nullptr;
 
+            if (auto focus = initial_focus_view.lock())
+            {
+                wf::get_core().default_wm->focus_raise_view(focus);
+            }
+
+            initial_focus_view.reset();
             return;
 
           default:
@@ -781,7 +772,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         auto views = get_views();
 
         return std::find(
-            views.begin(), views.end(), get_top_parent(view)) != views.end();
+            views.begin(), views.end(), wf::find_topmost_parent(view)) != views.end();
     }
 
     /* Convenience assignment function */
@@ -802,7 +793,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
             view_data.transformer->translation_y, translation_y);
         view_data.animation.scale_animation.start();
         view_data.fade_animation = wf::animation::simple_animation_t(
-            wf::option_wrapper_t<int>{"scale/duration"});
+            wf::option_wrapper_t<wf::animation_description_t>{"scale/duration"});
         view_data.fade_animation.animate(view_data.transformer->alpha,
             target_alpha);
     }
@@ -864,7 +855,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
          * transformer and are in scale_data */
         for (auto view : filtered_views)
         {
-            for (auto v : view->enumerate_views(false))
+            for (auto v : view->enumerate_views(true))
             {
                 add_transformer(v);
                 auto& view_data = scale_data[v];
@@ -907,6 +898,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
      * plugin algorithm */
     void layout_slots(std::vector<wayfire_toplevel_view> views)
     {
+        wf::dassert(active || hook_set, "Scale is not active");
         if (!views.size())
         {
             if (!all_workspaces && active)
@@ -984,7 +976,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
                 add_transformer(view);
                 auto geom = view->get_geometry();
                 double view_scale = calculate_scale({geom.width, geom.height});
-                for (auto& child : view->enumerate_views(false))
+                for (auto& child : view->enumerate_views(true))
                 {
                     // Ensure a transformer for the view, and make sure that
                     // new views in the view tree start off with the correct
@@ -1092,13 +1084,19 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         layout_slots(get_views());
     };
 
-    void handle_new_view(wayfire_toplevel_view view)
+    void handle_new_view(wayfire_toplevel_view view, bool close_scale)
     {
         if (!should_scale_view(view))
         {
             return;
         }
 
+        if (close_scale)
+        {
+            deactivate();
+            return;
+        }
+
         layout_slots(get_views());
     }
 
@@ -1106,53 +1104,22 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
     {
         if (auto toplevel = wf::toplevel_cast(ev->view))
         {
-            handle_new_view(toplevel);
+            handle_new_view(toplevel, close_on_new_view);
         }
     };
 
-    void handle_view_disappeared(wayfire_toplevel_view view)
+    void handle_view_unmapped(wayfire_toplevel_view view)
     {
-        if (scale_data.count(get_top_parent(view)) != 0)
+        remove_view(view);
+        if (scale_data.empty())
         {
-            /* Note: this signal can be emitted both when a view is removed from an output
-             * (closed or moved to another output) and when it is minimized. We use
-             * should_scale_view() to distinguish between the two cases. This only matters
-             * if we show minimized views. */
-            if (include_minimized && view->minimized && should_scale_view(view))
-            {
-                if (!scale_data.at(view).was_minimized)
-                {
-                    scale_data.at(view).was_minimized = true;
-                    wf::scene::set_node_enabled(view->get_root_node(), true);
-                }
-
-                fade_out(view);
-            } else
-            {
-                remove_view(view);
-                if (scale_data.empty())
-                {
-                    finalize();
-                }
-
-                if (!view->parent)
-                {
-                    layout_slots(get_views());
-                }
-            }
+            finalize();
+        } else if (!view->parent)
+        {
+            layout_slots(get_views());
         }
     }
 
-    /* Destroyed view or view moved to another output */
-    wf::signal::connection_t<wf::view_disappeared_signal> view_disappeared =
-        [=] (wf::view_disappeared_signal *ev)
-    {
-        if (auto toplevel = toplevel_cast(ev->view))
-        {
-            handle_view_disappeared(toplevel);
-        }
-    };
-
     /* Workspace changed */
     wf::signal::connection_t<wf::workspace_changed_signal> workspace_changed =
         [=] (wf::workspace_changed_signal *ev)
@@ -1193,6 +1160,15 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         if (!ev->view->minimized)
         {
             layout_slots(get_views());
+        } else if (include_minimized && scale_data.count(ev->view))
+        {
+            if (!scale_data.at(ev->view).was_minimized)
+            {
+                scale_data.at(ev->view).was_minimized = true;
+                wf::scene::set_node_enabled(ev->view->get_root_node(), true);
+            }
+
+            fade_out(ev->view);
         }
     };
 
@@ -1202,6 +1178,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         if (auto toplevel = wf::toplevel_cast(ev->view))
         {
             check_focus_view(toplevel);
+            handle_view_unmapped(toplevel);
         }
     };
 
@@ -1229,7 +1206,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
             }
         }
 
-        wf::get_core().default_wm->focus_raise_view(current_focus_view);
+        wf::get_core().default_wm->focus_raise_view(next_focus);
     }
 
     /* Returns true if any scale animation is running */
@@ -1334,9 +1311,23 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
             return false;
         }
 
-        initial_workspace  = output->wset()->get_current_workspace();
-        initial_focus_view = toplevel_cast(wf::get_active_view_for_output(output));
-        current_focus_view = initial_focus_view ?: views.front();
+        initial_workspace = output->wset()->get_current_workspace();
+
+        auto active_view = wf::get_active_view_for_output(output);
+        if (active_view)
+        {
+            initial_focus_view = active_view->weak_from_this();
+            current_focus_view = toplevel_cast(active_view);
+            if (std::find(views.begin(), views.end(), current_focus_view) == views.end())
+            {
+                current_focus_view = nullptr;
+            }
+        } else
+        {
+            initial_focus_view.reset();
+            current_focus_view = nullptr;
+        }
+
         // Make sure no leftover events from the activation binding
         // trigger an action in scale
         last_selected_view = nullptr;
@@ -1354,9 +1345,7 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         output->connect(&on_view_mapped);
         output->connect(&workspace_changed);
         output->connect(&workarea_changed);
-        output->connect(&view_disappeared);
         output->connect(&view_minimized);
-        output->connect(&view_unmapped);
 
         fade_out_all_except(current_focus_view);
         fade_in(current_focus_view);
@@ -1371,7 +1360,6 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
 
         set_hook();
         on_view_mapped.disconnect();
-        view_unmapped.disconnect();
         view_minimized.disconnect();
         workspace_changed.disconnect();
         workarea_changed.disconnect();
@@ -1427,8 +1415,6 @@ class wayfire_scale : public wf::per_output_plugin_instance_t,
         scale_data.clear();
         grab->ungrab_input();
         on_view_mapped.disconnect();
-        view_unmapped.disconnect();
-        view_disappeared.disconnect();
         view_minimized.disconnect();
         workspace_changed.disconnect();
         workarea_changed.disconnect();
@@ -1512,13 +1498,13 @@ class wayfire_scale_global : public wf::plugin_interface_t,
             auto old_output = ev->output;
             if (old_output && output_instance.count(old_output))
             {
-                this->output_instance[old_output]->handle_view_disappeared(toplevel);
+                this->output_instance[old_output]->handle_view_unmapped(toplevel);
             }
 
             auto new_output = ev->view->get_output();
             if (new_output && output_instance.count(new_output) && output_instance[new_output]->active)
             {
-                this->output_instance[ev->view->get_output()]->handle_new_view(toplevel);
+                this->output_instance[ev->view->get_output()]->handle_new_view(toplevel, false);
             }
         }
     };
