@@ -37,7 +37,7 @@ class wayfire_expo : public wf::per_output_plugin_instance_t, public wf::keyboar
     }
 
     wf::option_wrapper_t<wf::color_t> background_color{"expo/background"};
-    wf::option_wrapper_t<int> zoom_duration{"expo/duration"};
+    wf::option_wrapper_t<wf::animation_description_t> zoom_duration{"expo/duration"};
     wf::option_wrapper_t<int> delimiter_offset{"expo/offset"};
     wf::option_wrapper_t<bool> keyboard_interaction{"expo/keyboard_interaction"};
     wf::option_wrapper_t<double> inactive_brightness{"expo/inactive_brightness"};
@@ -372,11 +372,36 @@ class wayfire_expo : public wf::per_output_plugin_instance_t, public wf::keyboar
      */
     void handle_input_press(int32_t x, int32_t y, uint32_t state)
     {
-        if (zoom_animation.running() || !this->state.active)
+        const bool zooming = zoom_animation.running();
+
+        if (!this->state.active)
         {
             return;
         }
 
+        if (zooming && !this->state.zoom_in)
+        {
+            if (state == WLR_BUTTON_PRESSED)
+            {
+                // Zoom out of expo, change target workspace only
+                if (update_target_workspace(x, y))
+                {
+                    output->wset()->set_workspace(target_ws);
+
+                    // Calculate animation so that it appears as if we smoothly changed destination
+                    wf::geometry_t cur_viewport    = zoom_animation;
+                    wf::geometry_t target_viewport = wall->get_workspace_rectangle(target_ws);
+                    float cur_a = std::clamp(zoom_animation.progress(), 0.01, 0.99);
+                    wf::geometry_t start_viewport = wf::interpolate(cur_viewport, target_viewport,
+                        1.0 - 1.0 / (1.0 - cur_a));
+                    zoom_animation.set_start(start_viewport);
+                    zoom_animation.set_end(target_viewport);
+                }
+            }
+
+            return;
+        }
+
         if ((state == WLR_BUTTON_RELEASED) && !this->drag_helper->view)
         {
             this->state.button_pressed = false;
@@ -619,29 +644,49 @@ class wayfire_expo : public wf::per_output_plugin_instance_t, public wf::keyboar
     {
         auto local = input_coordinates_to_output_local_coordinates({gx, gy});
         wf::pointf_t localf = {1.0 * local.x, 1.0 * local.y};
-        return wf::find_output_view_at(output, localf);
+        auto view = wf::find_output_view_at(output, localf);
+        if (view && view->is_mapped())
+        {
+            return view;
+        } else
+        {
+            return nullptr;
+        }
     }
 
-    void update_target_workspace(int x, int y)
+    std::optional<wf::point_t> find_workspace_at(int x, int y)
     {
-        auto og = output->get_layout_geometry();
+        wlr_box box = {x, y, 1, 1};
+        auto og     = output->get_relative_geometry();
+        auto in_grid = wf::origin(wf::scale_box(og, wall->get_viewport(), box));
 
-        input_coordinates_to_global_coordinates(x, y);
-
-        auto grid = get_grid_geometry();
-        if (!(grid & wf::point_t{x, y}))
+        auto grid = output->wset()->get_workspace_grid_size();
+        for (int i = 0; i < grid.width; i++)
         {
-            return;
+            for (int j = 0; j < grid.height; j++)
+            {
+                if (wall->get_workspace_rectangle({i, j}) & in_grid)
+                {
+                    return wf::point_t{i, j};
+                }
+            }
         }
 
-        int tmpx = x / og.width;
-        int tmpy = y / og.height;
-        if ((tmpx != target_ws.x) || (tmpy != target_ws.y))
+        return std::nullopt;
+    }
+
+    bool update_target_workspace(int x, int y)
+    {
+        auto tmp = find_workspace_at(x, y);
+        if (tmp.has_value() && (tmp != target_ws))
         {
             shade_workspace(target_ws, true);
-            target_ws = {tmpx, tmpy};
+            target_ws = tmp.value();
             shade_workspace(target_ws, false);
+            return true;
         }
+
+        return false;
     }
 
     wf::effect_hook_t pre_frame = [=] ()
