@@ -1,21 +1,32 @@
-#include <wayfire/plugin.hpp>
+#include "wayfire/object.hpp"
+#include "wayfire/plugins/common/input-grab.hpp"
+#include "wayfire/scene-input.hpp"
+#include "wayfire/scene-operations.hpp"
+#include "wayfire/scene-render.hpp"
+#include "wayfire/scene.hpp"
+#include "wayfire/view-helpers.hpp"
+#include <wayfire/window-manager.hpp>
+#include <memory>
+#include <wayfire/per-output-plugin.hpp>
 #include <wayfire/opengl.hpp>
 #include <wayfire/view-transform.hpp>
 #include <wayfire/core.hpp>
+#include <wayfire/debug.hpp>
+#include <wayfire/plugins/common/util.hpp>
+#include <wayfire/seat.hpp>
 
 #include <wayfire/view.hpp>
 #include <wayfire/output.hpp>
 #include <wayfire/signal-definitions.hpp>
 
 #include <wayfire/render-manager.hpp>
-#include <wayfire/workspace-manager.hpp>
+#include <wayfire/workspace-set.hpp>
 
 #include <wayfire/util/duration.hpp>
 #include <wayfire/nonstd/reverse.hpp>
 #include <glm/gtc/matrix_transform.hpp>
 
 #include <algorithm>
-#include <exception>
 #include <set>
 
 constexpr const char *switcher_transformer = "switcher-3d";
@@ -52,7 +63,7 @@ static constexpr bool view_expired(int view_position)
 
 struct SwitcherView
 {
-    wayfire_view view;
+    wayfire_toplevel_view view;
     SwitcherPaintAttribs attribs;
 
     int position;
@@ -85,16 +96,20 @@ struct SwitcherView
     }
 };
 
-class WayfireSwitcher : public wf::plugin_interface_t
+class WayfireSwitcher : public wf::per_output_plugin_instance_t, public wf::keyboard_interaction_t
 {
     wf::option_wrapper_t<double> view_thumbnail_scale{
         "switcher/view_thumbnail_scale"};
-    wf::option_wrapper_t<int> speed{"switcher/speed"};
+    wf::option_wrapper_t<wf::animation_description_t> speed{"switcher/speed"};
+    wf::option_wrapper_t<int> view_thumbnail_rotation{
+        "switcher/view_thumbnail_rotation"};
 
     duration_t duration{speed};
     duration_t background_dim_duration{speed};
     timed_transition_t background_dim{background_dim_duration};
 
+    std::unique_ptr<wf::input_grab_t> input_grab;
+
     /* If a view comes before another in this list, it is on top of it */
     std::vector<SwitcherView> views;
 
@@ -102,56 +117,145 @@ class WayfireSwitcher : public wf::plugin_interface_t
     uint32_t activating_modifiers = 0;
     bool active = false;
 
-  public:
+    class switcher_render_node_t : public wf::scene::node_t
+    {
+        class switcher_render_instance_t : public wf::scene::render_instance_t
+        {
+            std::shared_ptr<switcher_render_node_t> self;
+            wf::scene::damage_callback push_damage;
+            wf::signal::connection_t<wf::scene::node_damage_signal> on_switcher_damage =
+                [=] (wf::scene::node_damage_signal *ev)
+            {
+                push_damage(ev->region);
+            };
 
+          public:
+            switcher_render_instance_t(switcher_render_node_t *self, wf::scene::damage_callback push_damage)
+            {
+                this->self = std::dynamic_pointer_cast<switcher_render_node_t>(self->shared_from_this());
+                this->push_damage = push_damage;
+                self->connect(&on_switcher_damage);
+            }
+
+            void schedule_instructions(
+                std::vector<wf::scene::render_instruction_t>& instructions,
+                const wf::render_target_t& target, wf::region_t& damage) override
+            {
+                instructions.push_back(wf::scene::render_instruction_t{
+                    .instance = this,
+                    .target   = target,
+                    .damage   = damage & self->get_bounding_box(),
+                });
+
+                // Don't render anything below
+                auto bbox = self->get_bounding_box();
+                damage ^= bbox;
+            }
+
+            void render(const wf::render_target_t& target,
+                const wf::region_t& region, const std::any& tag) override
+            {
+                self->switcher->render(target.translated(-wf::origin(self->get_bounding_box())));
+            }
+        };
+
+      public:
+        switcher_render_node_t(WayfireSwitcher *switcher) : node_t(false)
+        {
+            this->switcher = switcher;
+        }
+
+        virtual void gen_render_instances(
+            std::vector<wf::scene::render_instance_uptr>& instances,
+            wf::scene::damage_callback push_damage, wf::output_t *shown_on)
+        {
+            if (shown_on != this->switcher->output)
+            {
+                return;
+            }
+
+            instances.push_back(std::make_unique<switcher_render_instance_t>(this, push_damage));
+        }
+
+        wf::geometry_t get_bounding_box()
+        {
+            return switcher->output->get_layout_geometry();
+        }
+
+      private:
+        WayfireSwitcher *switcher;
+    };
+
+    std::shared_ptr<switcher_render_node_t> render_node;
+    wf::plugin_activation_data_t grab_interface = {
+        .name = "switcher",
+        .capabilities = wf::CAPABILITY_MANAGE_COMPOSITOR,
+    };
+
+  public:
     void init() override
     {
-        grab_interface->name = "switcher";
-        grab_interface->capabilities = wf::CAPABILITY_MANAGE_COMPOSITOR;
-
         output->add_key(
             wf::option_wrapper_t<wf::keybinding_t>{"switcher/next_view"},
             &next_view_binding);
         output->add_key(
             wf::option_wrapper_t<wf::keybinding_t>{"switcher/prev_view"},
             &prev_view_binding);
-        output->connect_signal("view-detached", &view_removed);
+        output->connect(&view_disappeared);
 
-        grab_interface->callbacks.keyboard.mod = [=] (uint32_t mod, uint32_t state)
+        input_grab = std::make_unique<wf::input_grab_t>("switcher", output, this, nullptr, nullptr);
+        grab_interface.cancel = [=] () {deinit_switcher();};
+    }
+
+    void handle_keyboard_key(wf::seat_t*, wlr_keyboard_key_event event) override
+    {
+        auto mod = wf::get_core().seat->modifier_from_keycode(event.keycode);
+        if ((event.state == WLR_KEY_RELEASED) && (mod & activating_modifiers))
         {
-            if ((state == WLR_KEY_RELEASED) && (mod & activating_modifiers))
-            {
-                handle_done();
-            }
-        };
-
-        grab_interface->callbacks.cancel = [=] () {deinit_switcher();};
+            handle_done();
+        }
     }
 
     wf::key_callback next_view_binding = [=] (auto)
     {
-        return handle_switch_request(-1);
+        handle_switch_request(-1);
+        return false;
     };
 
     wf::key_callback prev_view_binding = [=] (auto)
     {
-        return handle_switch_request(1);
+        handle_switch_request(1);
+        return false;
     };
 
-    wf::effect_hook_t damage = [=] ()
+    wf::effect_hook_t pre_hook = [=] ()
     {
-        output->render->damage_whole();
+        dim_background(background_dim);
+        wf::scene::damage_node(render_node, render_node->get_bounding_box());
+
+        if (!duration.running())
+        {
+            cleanup_expired();
+            if (!active)
+            {
+                deinit_switcher();
+            }
+        }
     };
 
-    wf::signal_callback_t view_removed = [=] (wf::signal_data_t *data)
+    wf::signal::connection_t<wf::view_disappeared_signal> view_disappeared =
+        [=] (wf::view_disappeared_signal *ev)
     {
-        handle_view_removed(get_signaled_view(data));
+        if (auto toplevel = toplevel_cast(ev->view))
+        {
+            handle_view_removed(toplevel);
+        }
     };
 
-    void handle_view_removed(wayfire_view view)
+    void handle_view_removed(wayfire_toplevel_view view)
     {
         // not running at all, don't care
-        if (!output->is_plugin_active(grab_interface->name))
+        if (!output->is_plugin_active(grab_interface.name))
         {
             return;
         }
@@ -186,7 +290,7 @@ class WayfireSwitcher : public wf::plugin_interface_t
         }
 
         /* If we haven't grabbed, then we haven't setup anything */
-        if (!output->is_plugin_active(grab_interface->name))
+        if (!output->is_plugin_active(grab_interface.name))
         {
             if (!init_switcher())
             {
@@ -199,14 +303,11 @@ class WayfireSwitcher : public wf::plugin_interface_t
         if (!active)
         {
             active = true;
-
-            // grabs shouldn't fail if we could successfully activate plugin
-            auto grab = grab_interface->grab();
-            assert(grab);
+            input_grab->grab_input(wf::scene::layer::OVERLAY);
 
             focus_next(dir);
             arrange();
-            activating_modifiers = wf::get_core().get_keyboard_modifiers();
+            activating_modifiers = wf::get_core().seat->get_keyboard_modifiers();
         } else
         {
             next_view(dir);
@@ -220,51 +321,57 @@ class WayfireSwitcher : public wf::plugin_interface_t
     {
         cleanup_expired();
         dearrange();
-        grab_interface->ungrab();
+        input_grab->ungrab_input();
     }
 
     /* Sets up basic hooks needed while switcher works and/or displays animations.
      * Also lower any fullscreen views that are active */
     bool init_switcher()
     {
-        if (!output->activate_plugin(grab_interface))
+        if (!output->activate_plugin(&grab_interface))
         {
             return false;
         }
 
-        output->render->add_effect(&damage, wf::OUTPUT_EFFECT_PRE);
-        output->render->set_renderer(switcher_renderer);
-        output->render->set_redraw_always();
+        output->render->add_effect(&pre_hook, wf::OUTPUT_EFFECT_PRE);
 
+        render_node = std::make_shared<switcher_render_node_t>(this);
+        wf::scene::add_front(wf::get_core().scene(), render_node);
         return true;
     }
 
     /* The reverse of init_switcher */
     void deinit_switcher()
     {
-        output->deactivate_plugin(grab_interface);
+        output->deactivate_plugin(&grab_interface);
 
-        output->render->rem_effect(&damage);
-        output->render->set_renderer(nullptr);
-        output->render->set_redraw_always(false);
+        output->render->rem_effect(&pre_hook);
+        wf::scene::remove_child(render_node);
+        render_node = nullptr;
 
-        for (auto& view : output->workspace->get_views_in_layer(wf::ALL_LAYERS))
+        for (auto& view : output->wset()->get_views())
         {
-            view->pop_transformer(switcher_transformer);
-            view->pop_transformer(switcher_transformer_background);
+            if (view->has_data("switcher-minimized-showed"))
+            {
+                view->erase_data("switcher-minimized-showed");
+                wf::scene::set_node_enabled(view->get_root_node(), false);
+            }
+
+            view->get_transformed_node()->rem_transformer(switcher_transformer);
+            view->get_transformed_node()->rem_transformer(
+                switcher_transformer_background);
         }
 
         views.clear();
 
-        wf::stack_order_changed_signal data;
-        data.output = output;
-        wf::get_core().emit_signal("output-stack-order-changed", &data);
+        wf::scene::update(wf::get_core().scene(),
+            wf::scene::update_flag::INPUT_STATE);
     }
 
     /* offset from the left or from the right */
     float get_center_offset()
     {
-        return output->get_relative_geometry().width / 3;
+        return output->get_relative_geometry().width / 3.0;
     }
 
     /* get the scale for non-focused views */
@@ -279,12 +386,6 @@ class WayfireSwitcher : public wf::plugin_interface_t
         return -1.0;
     }
 
-    /* amount of rotation */
-    float get_rotation()
-    {
-        return -M_PI / 6.0;
-    }
-
     /* Move view animation target to the left
      * @param dir -1 for left, 1 for right */
     void move(SwitcherView& sv, int dir)
@@ -318,8 +419,9 @@ class WayfireSwitcher : public wf::plugin_interface_t
         sv.attribs.scale_y.restart_with_end(
             sv.attribs.scale_y.end * std::pow(get_back_scale(), z_sign));
 
+        float radians_rotation = -((float)view_thumbnail_rotation * (M_PI / 180.0));
         sv.attribs.rotation.restart_with_end(
-            sv.attribs.rotation.end + get_rotation() * dir);
+            sv.attribs.rotation.end + radians_rotation * dir);
 
         sv.position += dir;
         sv.attribs.alpha.restart_with_end(
@@ -346,7 +448,7 @@ class WayfireSwitcher : public wf::plugin_interface_t
     }
 
     /* Calculate alpha for the view when switcher is inactive. */
-    float get_view_normal_alpha(wayfire_view view)
+    float get_view_normal_alpha(wayfire_toplevel_view view)
     {
         /* Usually views are visible, but if they were minimized,
          * and we aren't restoring the view, it has target alpha 0.0 */
@@ -362,10 +464,10 @@ class WayfireSwitcher : public wf::plugin_interface_t
     void arrange_center_view(SwitcherView& sv)
     {
         auto og   = output->get_relative_geometry();
-        auto bbox = sv.view->get_bounding_box(switcher_transformer);
+        auto bbox = wf::view_bounding_box_up_to(sv.view, switcher_transformer);
 
-        float dx = (og.width / 2 - bbox.width / 2) - bbox.x;
-        float dy = bbox.y - (og.height / 2 - bbox.height / 2);
+        float dx = (og.width / 2.0 - bbox.width / 2.0) - bbox.x;
+        float dy = bbox.y - (og.height / 2.0 - bbox.height / 2.0);
 
         sv.attribs.off_x.set(0, dx);
         sv.attribs.off_y.set(0, dy);
@@ -391,22 +493,9 @@ class WayfireSwitcher : public wf::plugin_interface_t
     }
 
     // returns a list of mapped views
-    std::vector<wayfire_view> get_workspace_views() const
+    std::vector<wayfire_toplevel_view> get_workspace_views() const
     {
-        auto all_views = output->workspace->get_views_on_workspace(
-            output->workspace->get_current_workspace(),
-            wf::WM_LAYERS | wf::LAYER_MINIMIZED);
-
-        decltype(all_views) mapped_views;
-        for (auto view : all_views)
-        {
-            if (view->is_mapped())
-            {
-                mapped_views.push_back(view);
-            }
-        }
-
-        return mapped_views;
+        return output->wset()->get_views(wf::WSET_MAPPED_ONLY | wf::WSET_CURRENT_WORKSPACE);
     }
 
     /* Change the current focus to the next or the previous view */
@@ -419,8 +508,7 @@ class WayfireSwitcher : public wf::plugin_interface_t
         // calculate focus index & focus it
         int focused_view_index = (size + dir) % size;
         auto focused_view = ws_views[focused_view_index];
-
-        output->workspace->bring_to_front(focused_view);
+        wf::view_bring_to_front(focused_view);
     }
 
     /* Create the initial arrangement on the screen
@@ -440,6 +528,16 @@ class WayfireSwitcher : public wf::plugin_interface_t
             views.push_back(create_switcher_view(v));
         }
 
+        std::sort(views.begin(), views.end(), [] (SwitcherView& a, SwitcherView& b)
+        {
+            return wf::get_focus_timestamp(a.view) > wf::get_focus_timestamp(b.view);
+        });
+
+        if (ws_views.empty())
+        {
+            return;
+        }
+
         /* Add a copy of the unfocused view if we have just 2 */
         if (ws_views.size() == 2)
         {
@@ -458,6 +556,10 @@ class WayfireSwitcher : public wf::plugin_interface_t
         {
             arrange_view(views[i], SWITCHER_POSITION_RIGHT);
         }
+
+        // We want the next view to be focused right off the bat
+        // But we want it to be animated.
+        handle_switch_request(-1);
     }
 
     void dearrange()
@@ -466,7 +568,7 @@ class WayfireSwitcher : public wf::plugin_interface_t
          * of the unfocused view. When dearranging those copies, they overlap.
          * If the view is translucent, this means that the view gets darker than
          * it really is.  * To circumvent this, we just fade out one of the copies */
-        wayfire_view fading_view = nullptr;
+        wayfire_toplevel_view fading_view = nullptr;
         if (count_different_active_views() == 2)
         {
             fading_view = get_unfocused_view();
@@ -501,20 +603,20 @@ class WayfireSwitcher : public wf::plugin_interface_t
         /* Potentially restore view[0] if it was maximized */
         if (views.size())
         {
-            output->focus_view(views[0].view, true);
+            wf::get_core().default_wm->focus_raise_view(views[0].view);
         }
     }
 
     std::vector<wayfire_view> get_background_views() const
     {
-        return output->workspace->get_views_on_workspace(
-            output->workspace->get_current_workspace(), wf::BELOW_LAYERS);
+        return wf::collect_views_from_output(output,
+            {wf::scene::layer::BACKGROUND, wf::scene::layer::BOTTOM});
     }
 
     std::vector<wayfire_view> get_overlay_views() const
     {
-        return output->workspace->get_views_on_workspace(
-            output->workspace->get_current_workspace(), wf::ABOVE_LAYERS);
+        return wf::collect_views_from_output(output,
+            {wf::scene::layer::TOP, wf::scene::layer::OVERLAY, wf::scene::layer::DWIDGET});
     }
 
     void dim_background(float dim)
@@ -523,33 +625,38 @@ class WayfireSwitcher : public wf::plugin_interface_t
         {
             if (dim == 1.0)
             {
-                view->pop_transformer(switcher_transformer_background);
+                view->get_transformed_node()->rem_transformer(
+                    switcher_transformer_background);
             } else
             {
-                if (!view->get_transformer(switcher_transformer_background))
-                {
-                    view->add_transformer(std::make_unique<wf::view_3D>(view),
-                        switcher_transformer_background);
-                }
-
-                auto tr = dynamic_cast<wf::view_3D*>(
-                    view->get_transformer(switcher_transformer_background).get());
+                auto tr =
+                    wf::ensure_named_transformer<wf::scene::view_3d_transformer_t>(
+                        view, wf::TRANSFORMER_3D, switcher_transformer_background,
+                        view);
                 tr->color[0] = tr->color[1] = tr->color[2] = dim;
             }
         }
     }
 
-    SwitcherView create_switcher_view(wayfire_view view)
+    SwitcherView create_switcher_view(wayfire_toplevel_view view)
     {
         /* we add a view transform if there isn't any.
          *
          * Note that a view might be visible on more than 1 place, so damage
          * tracking doesn't work reliably. To circumvent this, we simply damage
          * the whole output */
-        if (!view->get_transformer(switcher_transformer))
+        if (!view->get_transformed_node()->get_transformer(switcher_transformer))
         {
-            view->add_transformer(std::make_unique<wf::view_3D>(view),
-                switcher_transformer);
+            if (view->minimized)
+            {
+                wf::scene::set_node_enabled(view->get_root_node(), true);
+                view->store_data(std::make_unique<wf::custom_data_t>(),
+                    "switcher-minimized-showed");
+            }
+
+            view->get_transformed_node()->add_transformer(
+                std::make_shared<wf::scene::view_3d_transformer_t>(view),
+                wf::TRANSFORMER_3D, switcher_transformer);
         }
 
         SwitcherView sw{duration};
@@ -559,10 +666,23 @@ class WayfireSwitcher : public wf::plugin_interface_t
         return sw;
     }
 
-    void render_view(const SwitcherView& sv, const wf::framebuffer_t& buffer)
+    void render_view_scene(wayfire_view view, const wf::render_target_t& buffer)
     {
-        auto transform = dynamic_cast<wf::view_3D*>(
-            sv.view->get_transformer(switcher_transformer).get());
+        std::vector<wf::scene::render_instance_uptr> instances;
+        view->get_transformed_node()->gen_render_instances(instances, [] (auto) {});
+
+        wf::scene::render_pass_params_t params;
+        params.instances = &instances;
+        params.damage    = view->get_transformed_node()->get_bounding_box();
+        params.reference_output = this->output;
+        params.target = buffer;
+        wf::scene::run_render_pass(params, 0);
+    }
+
+    void render_view(const SwitcherView& sv, const wf::render_target_t& buffer)
+    {
+        auto transform = sv.view->get_transformed_node()
+            ->get_transformer<wf::scene::view_3d_transformer_t>(switcher_transformer);
         assert(transform);
 
         transform->translation = glm::translate(glm::mat4(1.0),
@@ -576,19 +696,18 @@ class WayfireSwitcher : public wf::plugin_interface_t
             (float)sv.attribs.rotation, {0.0, 1.0, 0.0});
 
         transform->color[3] = sv.attribs.alpha;
-        sv.view->render_transformed(buffer, buffer.geometry);
+        render_view_scene(sv.view, buffer);
     }
 
-    wf::render_hook_t switcher_renderer = [=] (const wf::framebuffer_t& fb)
+    void render(const wf::render_target_t& fb)
     {
         OpenGL::render_begin(fb);
         OpenGL::clear({0, 0, 0, 1});
         OpenGL::render_end();
 
-        dim_background(background_dim);
         for (auto view : get_background_views())
         {
-            view->render_transformed(fb, fb.geometry);
+            render_view_scene(view, fb);
         }
 
         /* Render in the reverse order because we don't use depth testing */
@@ -599,19 +718,9 @@ class WayfireSwitcher : public wf::plugin_interface_t
 
         for (auto view : get_overlay_views())
         {
-            view->render_transformed(fb, fb.geometry);
+            render_view_scene(view, fb);
         }
-
-        if (!duration.running())
-        {
-            cleanup_expired();
-
-            if (!active)
-            {
-                deinit_switcher();
-            }
-        }
-    };
+    }
 
     /* delete all views matching the given criteria, skipping the first "start" views
      * */
@@ -713,17 +822,17 @@ class WayfireSwitcher : public wf::plugin_interface_t
         if (bool(count_left) ^ bool(count_right))
         {
             const int empty_slot = 1 - dir;
-            fill_emtpy_slot(empty_slot);
+            fill_empty_slot(empty_slot);
         }
 
         rebuild_view_list();
-        output->workspace->bring_to_front(views.front().view);
+        wf::view_bring_to_front(views.front().view);
         duration.start();
     }
 
     int count_different_active_views()
     {
-        std::set<wayfire_view> active_views;
+        std::set<wayfire_toplevel_view> active_views;
         for (auto& sv : views)
         {
             active_views.insert(sv.view);
@@ -733,7 +842,7 @@ class WayfireSwitcher : public wf::plugin_interface_t
     }
 
 /* Move the last view in the given slot so that it becomes invalid */
-    wayfire_view invalidate_last_in_slot(int slot)
+    wayfire_toplevel_view invalidate_last_in_slot(int slot)
     {
         for (int i = views.size() - 1; i >= 0; i--)
         {
@@ -749,7 +858,7 @@ class WayfireSwitcher : public wf::plugin_interface_t
     }
 
 /* Returns the non-focused view in the case where there is only 1 view */
-    wayfire_view get_unfocused_view()
+    wayfire_toplevel_view get_unfocused_view()
     {
         for (auto& sv : views)
         {
@@ -763,7 +872,7 @@ class WayfireSwitcher : public wf::plugin_interface_t
         return nullptr;
     }
 
-    void fill_emtpy_slot(const int empty_slot)
+    void fill_empty_slot(const int empty_slot)
     {
         const int full_slot = 2 - empty_slot;
 
@@ -794,15 +903,15 @@ class WayfireSwitcher : public wf::plugin_interface_t
 
     void fini() override
     {
-        if (output->is_plugin_active(grab_interface->name))
+        if (output->is_plugin_active(grab_interface.name))
         {
+            input_grab->ungrab_input();
             deinit_switcher();
         }
 
         output->rem_binding(&next_view_binding);
         output->rem_binding(&prev_view_binding);
-        output->disconnect_signal("view-detached", &view_removed);
     }
 };
 
-DECLARE_WAYFIRE_PLUGIN(WayfireSwitcher);
+DECLARE_WAYFIRE_PLUGIN(wf::per_output_plugin_t<WayfireSwitcher>);
