@@ -1,25 +1,33 @@
-#include <wayfire/plugin.hpp>
+#include <wayfire/per-output-plugin.hpp>
 #include <wayfire/output.hpp>
 #include <wayfire/opengl.hpp>
 #include <wayfire/render-manager.hpp>
 #include <wayfire/util/duration.hpp>
 
-class wayfire_zoom_screen : public wf::plugin_interface_t
+class wayfire_zoom_screen : public wf::per_output_plugin_instance_t
 {
+    enum class interpolation_method_t
+    {
+        LINEAR  = 0,
+        NEAREST = 1,
+    };
+
     wf::option_wrapper_t<wf::keybinding_t> modifier{"zoom/modifier"};
     wf::option_wrapper_t<double> speed{"zoom/speed"};
-    wf::option_wrapper_t<int> smoothing_duration{"zoom/smoothing_duration"};
+    wf::option_wrapper_t<wf::animation_description_t> smoothing_duration{"zoom/smoothing_duration"};
+    wf::option_wrapper_t<int> interpolation_method{"zoom/interpolation_method"};
     wf::animation::simple_animation_t progression{smoothing_duration};
     bool hook_set = false;
 
+    wf::plugin_activation_data_t grab_interface = {
+        .name = "zoom",
+        .capabilities = 0,
+    };
+
   public:
     void init() override
     {
-        grab_interface->name = "zoom";
-        grab_interface->capabilities = 0;
-
         progression.set(1, 1);
-
         output->add_axis(modifier, &axis);
     }
 
@@ -44,7 +52,7 @@ class wayfire_zoom_screen : public wf::plugin_interface_t
 
     wf::axis_callback axis = [=] (wlr_pointer_axis_event *ev)
     {
-        if (!output->can_activate_plugin(grab_interface))
+        if (!output->can_activate_plugin(&grab_interface))
         {
             return false;
         }
@@ -59,8 +67,8 @@ class wayfire_zoom_screen : public wf::plugin_interface_t
         return true;
     };
 
-    wf::post_hook_t render_hook = [=] (const wf::framebuffer_base_t& source,
-                                       const wf::framebuffer_base_t& destination)
+    wf::post_hook_t render_hook = [=] (const wf::framebuffer_t& source,
+                                       const wf::framebuffer_t& destination)
     {
         auto w = destination.viewport_width;
         auto h = destination.viewport_height;
@@ -87,11 +95,15 @@ class wayfire_zoom_screen : public wf::plugin_interface_t
         const float x1 = x * scale;
         const float y1 = y * scale;
 
+        const GLenum interpolation =
+            (interpolation_method ==
+                (int)interpolation_method_t::NEAREST) ? GL_NEAREST : GL_LINEAR;
+
         OpenGL::render_begin(source);
         GL_CALL(glBindFramebuffer(GL_READ_FRAMEBUFFER, source.fb));
         GL_CALL(glBindFramebuffer(GL_DRAW_FRAMEBUFFER, destination.fb));
         GL_CALL(glBlitFramebuffer(x1, y1, x1 + tw, y1 + th, 0, 0, w, h,
-            GL_COLOR_BUFFER_BIT, GL_LINEAR));
+            GL_COLOR_BUFFER_BIT, interpolation));
         OpenGL::render_end();
 
         if (!progression.running() && (progression - 1 <= 0.01))
@@ -118,4 +130,4 @@ class wayfire_zoom_screen : public wf::plugin_interface_t
     }
 };
 
-DECLARE_WAYFIRE_PLUGIN(wayfire_zoom_screen);
+DECLARE_WAYFIRE_PLUGIN(wf::per_output_plugin_t<wayfire_zoom_screen>);
