@@ -1,15 +1,14 @@
 #include <memory>
 #include <wayfire/config/types.hpp>
 #include <wayfire/per-output-plugin.hpp>
-#include <wayfire/workarea.hpp>
 #include <wayfire/output.hpp>
 #include <wayfire/core.hpp>
 #include <wayfire/matcher.hpp>
 #include <wayfire/workspace-set.hpp>
 #include <wayfire/signal-definitions.hpp>
 #include <wayfire/toplevel-view.hpp>
-#include <wayfire/window-manager.hpp>
 
+#include "plugins/ipc/ipc-method-repository.hpp"
 #include "tree-controller.hpp"
 #include "tree.hpp"
 #include "wayfire/debug.hpp"
@@ -18,35 +17,17 @@
 #include "wayfire/option-wrapper.hpp"
 #include "wayfire/plugin.hpp"
 #include "wayfire/plugins/common/input-grab.hpp"
+#include "wayfire/plugins/common/shared-core-data.hpp"
 #include "wayfire/scene-input.hpp"
-#include "wayfire/scene-operations.hpp"
 #include "wayfire/scene.hpp"
 #include "wayfire/signal-provider.hpp"
 #include "wayfire/toplevel-view.hpp"
-#include "wayfire/view-helpers.hpp"
-#include "wayfire/txn/transaction-manager.hpp"
 #include "wayfire/view.hpp"
 
-struct autocommit_transaction_t
-{
-  public:
-    wf::txn::transaction_uptr tx;
-    autocommit_transaction_t()
-    {
-        tx = wf::txn::transaction_t::create();
-    }
+#include "tile-wset.hpp"
+#include "tile-ipc.hpp"
+#include "tile-dragging.hpp"
 
-    ~autocommit_transaction_t()
-    {
-        if (!tx->get_objects().empty())
-        {
-            wf::get_core().tx_manager->schedule_transaction(std::move(tx));
-        }
-    }
-};
-
-namespace wf
-{
 static bool can_tile_view(wayfire_toplevel_view view)
 {
     if (view->parent)
@@ -54,276 +35,17 @@ static bool can_tile_view(wayfire_toplevel_view view)
         return false;
     }
 
+    if ((view->toplevel()->get_min_size() == view->toplevel()->get_max_size()) &&
+        (view->toplevel()->get_min_size().width > 0) && (view->toplevel()->get_min_size().height > 0))
+    {
+        return false;
+    }
+
     return true;
 }
 
-/**
- * When a view is moved from one output to the other, we want to keep its tiled
- * status. To achieve this, we do the following:
- *
- * 1. In view-pre-moved-to-output handler, we set view_auto_tile_t custom data.
- * 2. In detach handler, we just remove the view as usual.
- * 3. We now know we will receive attach as next event.
- *    Check for view_auto_tile_t, and tile the view again.
- */
-class view_auto_tile_t : public wf::custom_data_t
-{};
-
-class tile_workspace_set_data_t : public wf::custom_data_t
+namespace wf
 {
-  public:
-    std::vector<std::vector<std::unique_ptr<wf::tile::tree_node_t>>> roots;
-    std::vector<std::vector<wf::scene::floating_inner_ptr>> tiled_sublayer;
-    const wf::tile::split_direction_t default_split = wf::tile::SPLIT_VERTICAL;
-
-    wf::option_wrapper_t<int> inner_gaps{"simple-tile/inner_gap_size"};
-    wf::option_wrapper_t<int> outer_horiz_gaps{"simple-tile/outer_horiz_gap_size"};
-    wf::option_wrapper_t<int> outer_vert_gaps{"simple-tile/outer_vert_gap_size"};
-
-    tile_workspace_set_data_t(std::shared_ptr<wf::workspace_set_t> wset)
-    {
-        this->wset = wset;
-        wset->connect(&on_wset_attached);
-        wset->connect(&on_workspace_grid_changed);
-        resize_roots(wset->get_workspace_grid_size());
-        if (wset->get_attached_output())
-        {
-            wset->get_attached_output()->connect(&on_workarea_changed);
-        }
-
-        inner_gaps.set_callback(update_gaps);
-        outer_horiz_gaps.set_callback(update_gaps);
-        outer_vert_gaps.set_callback(update_gaps);
-    }
-
-    wf::signal::connection_t<workarea_changed_signal> on_workarea_changed = [=] (auto)
-    {
-        update_root_size();
-    };
-
-    wf::signal::connection_t<workspace_set_attached_signal> on_wset_attached = [=] (auto)
-    {
-        on_workarea_changed.disconnect();
-        if (wset.lock()->get_attached_output())
-        {
-            wset.lock()->get_attached_output()->connect(&on_workarea_changed);
-            update_root_size();
-        }
-    };
-
-    wf::signal::connection_t<wf::workspace_grid_changed_signal> on_workspace_grid_changed = [=] (auto)
-    {
-        wf::dassert(!wset.expired(), "wset should not expire, ever!");
-        resize_roots(wset.lock()->get_workspace_grid_size());
-    };
-
-    void resize_roots(wf::dimensions_t wsize)
-    {
-        for (size_t i = 0; i < tiled_sublayer.size(); i++)
-        {
-            for (size_t j = 0; j < tiled_sublayer[i].size(); j++)
-            {
-                if (wset.lock()->is_workspace_valid({(int)i, (int)j}))
-                {
-                    destroy_sublayer(tiled_sublayer[i][j]);
-                }
-            }
-        }
-
-        roots.resize(wsize.width);
-        tiled_sublayer.resize(wsize.width);
-        for (int i = 0; i < wsize.width; i++)
-        {
-            roots[i].resize(wsize.height);
-            tiled_sublayer[i].resize(wsize.height);
-            for (int j = 0; j < wsize.height; j++)
-            {
-                roots[i][j] = std::make_unique<wf::tile::split_node_t>(default_split);
-                tiled_sublayer[i][j] = std::make_shared<wf::scene::floating_inner_node_t>(false);
-                wf::scene::add_front(wset.lock()->get_node(), tiled_sublayer[i][j]);
-            }
-        }
-
-        update_root_size();
-        update_gaps();
-    }
-
-    void update_root_size()
-    {
-        auto wo = wset.lock()->get_attached_output();
-        wf::geometry_t workarea = wo ? wo->workarea->get_workarea() : tile::default_output_resolution;
-
-        wf::geometry_t output_geometry =
-            wset.lock()->get_last_output_geometry().value_or(tile::default_output_resolution);
-
-        auto wsize = wset.lock()->get_workspace_grid_size();
-        for (int i = 0; i < wsize.width; i++)
-        {
-            for (int j = 0; j < wsize.height; j++)
-            {
-                /* Set size */
-                auto vp_geometry = workarea;
-                vp_geometry.x += i * output_geometry.width;
-                vp_geometry.y += j * output_geometry.height;
-
-                autocommit_transaction_t tx;
-                roots[i][j]->set_geometry(vp_geometry, tx.tx);
-            }
-        }
-    }
-
-    void destroy_sublayer(wf::scene::floating_inner_ptr sublayer)
-    {
-        // Transfer views to the top
-        auto root     = wset.lock()->get_node();
-        auto children = root->get_children();
-        auto sublayer_children = sublayer->get_children();
-        sublayer->set_children_list({});
-        children.insert(children.end(), sublayer_children.begin(), sublayer_children.end());
-        root->set_children_list(children);
-        wf::scene::update(root, wf::scene::update_flag::CHILDREN_LIST);
-        wf::scene::remove_child(sublayer);
-    }
-
-    std::function<void()> update_gaps = [=] ()
-    {
-        tile::gap_size_t gaps = {
-            .left   = outer_horiz_gaps,
-            .right  = outer_horiz_gaps,
-            .top    = outer_vert_gaps,
-            .bottom = outer_vert_gaps,
-            .internal = inner_gaps,
-        };
-
-        for (auto& col : roots)
-        {
-            for (auto& root : col)
-            {
-                autocommit_transaction_t tx;
-                root->set_gaps(gaps, tx.tx);
-                root->set_geometry(root->geometry, tx.tx);
-            }
-        }
-    };
-
-    void flatten_roots()
-    {
-        for (auto& col : roots)
-        {
-            for (auto& root : col)
-            {
-                autocommit_transaction_t tx;
-                tile::flatten_tree(root, tx.tx);
-            }
-        }
-    }
-
-    static tile_workspace_set_data_t& get(std::shared_ptr<workspace_set_t> set)
-    {
-        if (!set->has_data<tile_workspace_set_data_t>())
-        {
-            set->store_data(std::make_unique<tile_workspace_set_data_t>(set));
-        }
-
-        return *set->get_data<tile_workspace_set_data_t>();
-    }
-
-    static tile_workspace_set_data_t& get(wf::output_t *output)
-    {
-        return get(output->wset());
-    }
-
-    static std::unique_ptr<tile::tree_node_t>& get_current_root(wf::output_t *output)
-    {
-        auto set   = output->wset();
-        auto vp    = set->get_current_workspace();
-        auto& data = get(output);
-        return data.roots[vp.x][vp.y];
-    }
-
-    static scene::floating_inner_ptr get_current_sublayer(wf::output_t *output)
-    {
-        auto set   = output->wset();
-        auto vp    = set->get_current_workspace();
-        auto& data = get(output);
-        return data.tiled_sublayer[vp.x][vp.y];
-    }
-
-    std::weak_ptr<workspace_set_t> wset;
-
-    void attach_view(wayfire_toplevel_view view, wf::point_t vp = {-1, -1})
-    {
-        view->set_allowed_actions(VIEW_ALLOW_WS_CHANGE);
-
-        if (vp == wf::point_t{-1, -1})
-        {
-            vp = wset.lock()->get_current_workspace();
-        }
-
-        auto view_node = std::make_unique<wf::tile::view_node_t>(view);
-        {
-            autocommit_transaction_t tx;
-            roots[vp.x][vp.y]->as_split_node()->add_child(std::move(view_node), tx.tx);
-        }
-
-        auto node = view->get_root_node();
-        wf::scene::readd_front(tiled_sublayer[vp.x][vp.y], node);
-        view_bring_to_front(view);
-        consider_exit_fullscreen(view);
-    }
-
-    /** Remove the given view from its tiling container */
-    void detach_view(nonstd::observer_ptr<tile::view_node_t> view,
-        bool reinsert = true)
-    {
-        auto wview = view->view;
-        wview->set_allowed_actions(VIEW_ALLOW_ALL);
-        {
-            autocommit_transaction_t tx;
-            view->parent->remove_child(view, tx.tx);
-        }
-        /* View node is invalid now */
-        flatten_roots();
-
-        if (wview->pending_fullscreen() && wview->is_mapped())
-        {
-            wf::get_core().default_wm->fullscreen_request(wview, nullptr, false);
-        }
-
-        /* Remove from special sublayer */
-        if (reinsert)
-        {
-            wf::scene::readd_front(wview->get_output()->wset()->get_node(), wview->get_root_node());
-        }
-    }
-
-    /**
-     * Consider unfullscreening all fullscreen views because a new view has been focused or attached to the
-     * tiling tree.
-     */
-    void consider_exit_fullscreen(wayfire_toplevel_view view)
-    {
-        if (tile::view_node_t::get_node(view) && !view->pending_fullscreen())
-        {
-            auto vp = this->wset.lock()->get_current_workspace();
-            for_each_view(roots[vp.x][vp.y], [&] (wayfire_toplevel_view view)
-            {
-                if (view->pending_fullscreen())
-                {
-                    set_view_fullscreen(view, false);
-                }
-            });
-        }
-    }
-
-    void set_view_fullscreen(wayfire_toplevel_view view, bool fullscreen)
-    {
-        /* Set fullscreen, and trigger resizing of the views (which will commit the view) */
-        view->toplevel()->pending().fullscreen = fullscreen;
-        update_root_size();
-    }
-};
-
 class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom_data_t
 {
   private:
@@ -350,22 +72,6 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
     std::unique_ptr<wf::tile::tile_controller_t> controller =
         get_default_controller();
 
-    /**
-     * Translate coordinates from output-local coordinates to the coordinate
-     * system of the tiling trees, depending on the current workspace
-     */
-    wf::point_t get_global_input_coordinates()
-    {
-        wf::pointf_t local = output->get_cursor_position();
-
-        auto vp   = output->wset()->get_current_workspace();
-        auto size = output->get_screen_size();
-        local.x += size.width * vp.x;
-        local.y += size.height * vp.y;
-
-        return {(int)local.x, (int)local.y};
-    }
-
     /** Check whether we currently have a fullscreen tiled view */
     bool has_fullscreen_view()
     {
@@ -379,18 +85,24 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
     }
 
     /** Check whether the current pointer focus is tiled view */
-    bool has_tiled_focus()
+    wayfire_toplevel_view get_tiled_focus()
     {
-        auto focus = wf::get_core().get_cursor_focus_view();
+        auto focus = toplevel_cast(wf::get_core().get_cursor_focus_view());
+        if (focus && tile::view_node_t::get_node(focus))
+        {
+            return focus;
+        }
 
-        return focus && tile::view_node_t::get_node(focus);
+        return nullptr;
     }
 
     template<class Controller>
     void start_controller()
     {
+        auto tiled_focus = get_tiled_focus();
+
         /* No action possible in this case */
-        if (has_fullscreen_view() || !has_tiled_focus())
+        if (has_fullscreen_view() || !tiled_focus)
         {
             return;
         }
@@ -401,8 +113,7 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
         }
 
         input_grab->grab_input(wf::scene::layer::OVERLAY);
-        controller = std::make_unique<Controller>(tile_workspace_set_data_t::get_current_root(output),
-            get_global_input_coordinates());
+        controller = std::make_unique<Controller>(output->wset().get(), tiled_focus);
     }
 
     void stop_controller(bool force_stop)
@@ -412,15 +123,10 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
             return;
         }
 
-        input_grab->ungrab_input();
-
         // Deactivate plugin, so that others can react to the events
         output->deactivate_plugin(&grab_interface);
-        if (!force_stop)
-        {
-            controller->input_released();
-        }
-
+        input_grab->ungrab_input();
+        controller->input_released(force_stop);
         controller = get_default_controller();
     }
 
@@ -429,7 +135,7 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
         return tile_by_default.matches(view) && can_tile_view(view);
     }
 
-    void attach_view(wayfire_toplevel_view view, wf::point_t vp = {-1, -1})
+    void attach_view(wayfire_toplevel_view view, std::optional<wf::point_t> vp = {})
     {
         if (!view->get_wset())
         {
@@ -440,10 +146,15 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
         tile_workspace_set_data_t::get(view->get_wset()).attach_view(view, vp);
     }
 
-    void detach_view(nonstd::observer_ptr<tile::view_node_t> view, bool reinsert = true)
+    void detach_view(wayfire_toplevel_view view, bool reinsert = true)
     {
         stop_controller(true);
-        tile_workspace_set_data_t::get(view->view->get_wset()).detach_view(view, reinsert);
+
+        // Note that stopping the controller might untile the view, or change its tiled node.
+        if (auto node = tile::view_node_t::get_node(view))
+        {
+            tile_workspace_set_data_t::get(view->get_wset()).detach_views({node}, reinsert);
+        }
     }
 
     wf::signal::connection_t<view_mapped_signal> on_view_mapped = [=] (view_mapped_signal *ev)
@@ -457,16 +168,6 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
         }
     };
 
-    wf::signal::connection_t<view_unmapped_signal> on_view_unmapped = [=] (wf::view_unmapped_signal *ev)
-    {
-        stop_controller(true);
-        auto node = wf::tile::view_node_t::get_node(ev->view);
-        if (node)
-        {
-            detach_view(node);
-        }
-    };
-
     wf::signal::connection_t<view_tile_request_signal> on_tile_request = [=] (view_tile_request_signal *ev)
     {
         if (ev->carried_out || !tile::view_node_t::get_node(ev->view))
@@ -490,12 +191,12 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
         tile_workspace_set_data_t::get(ev->view->get_wset()).set_view_fullscreen(ev->view, ev->state);
     };
 
-    void change_view_workspace(wayfire_toplevel_view view, wf::point_t vp = {-1, -1})
+    void change_view_workspace(wayfire_toplevel_view view, std::optional<wf::point_t> vp = {})
     {
         auto existing_node = wf::tile::view_node_t::get_node(view);
         if (existing_node)
         {
-            detach_view(existing_node);
+            detach_view(view);
             attach_view(view, vp);
         }
     }
@@ -515,7 +216,7 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
 
         if (ev->view->minimized && existing_node)
         {
-            detach_view(existing_node);
+            detach_view(ev->view);
         }
 
         if (!ev->view->minimized && tile_window_by_default(ev->view))
@@ -561,7 +262,7 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
             auto existing_node = tile::view_node_t::get_node(view);
             if (existing_node)
             {
-                detach_view(existing_node);
+                detach_view(view);
                 wf::get_core().default_wm->tile_request(view, 0);
             } else
             {
@@ -637,9 +338,9 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
         }
     }
 
-    void handle_pointer_motion(wf::pointf_t pointer_position, uint32_t time_ms) override
+    void handle_pointer_motion(wf::pointf_t, uint32_t) override
     {
-        controller->input_motion(get_global_input_coordinates());
+        controller->input_motion();
     }
 
     void setup_callbacks()
@@ -657,6 +358,7 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
     wf::plugin_activation_data_t grab_interface = {
         .name = "simple-tile",
         .capabilities = CAPABILITY_MANAGE_COMPOSITOR,
+        .cancel = [=] { stop_controller(true); },
     };
 
   public:
@@ -665,7 +367,6 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
         this->output = wo;
         input_grab   = std::make_unique<wf::input_grab_t>("simple-tile", output, nullptr, this, nullptr);
         output->connect(&on_view_mapped);
-        output->connect(&on_view_unmapped);
         output->connect(&on_tile_request);
         output->connect(&on_fullscreen_request);
         output->connect(&on_view_change_workspace);
@@ -679,11 +380,16 @@ class tile_output_plugin_t : public wf::pointer_interaction_t, public wf::custom
         output->rem_binding(&on_resize_view);
         output->rem_binding(&on_toggle_tiled_state);
         output->rem_binding(&on_focus_adjacent);
+        stop_controller(true);
     }
 };
 
 class tile_plugin_t : public wf::plugin_interface_t, wf::per_output_tracker_mixin_t<>
 {
+    shared_data::ref_ptr_t<ipc::method_repository_t> ipc_repo;
+    shared_data::ref_ptr_t<wf::move_drag::core_drag_t> drag_helper;
+    std::unique_ptr<tile::drag_manager_t> preview_manager;
+
   public:
     void init() override
     {
@@ -691,10 +397,15 @@ class tile_plugin_t : public wf::plugin_interface_t, wf::per_output_tracker_mixi
         wf::get_core().connect(&on_view_pre_moved_to_wset);
         wf::get_core().connect(&on_view_moved_to_wset);
         wf::get_core().connect(&on_focus_changed);
+        wf::get_core().connect(&on_view_unmapped);
+        ipc_repo->register_method("simple-tile/get-layout", ipc_get_layout);
+        ipc_repo->register_method("simple-tile/set-layout", ipc_set_layout);
+        preview_manager = std::make_unique<tile::drag_manager_t>();
     }
 
     void fini() override
     {
+        preview_manager.reset();
         fini_output_tracking();
         for (auto wset : workspace_set_t::get_all())
         {
@@ -705,6 +416,9 @@ class tile_plugin_t : public wf::plugin_interface_t, wf::per_output_tracker_mixi
         {
             wo->erase_data<tile_output_plugin_t>();
         }
+
+        ipc_repo->unregister_method("simple-tile/get-layout");
+        ipc_repo->unregister_method("simple-tile/set-layout");
     }
 
     void stop_controller(std::shared_ptr<wf::workspace_set_t> wset)
@@ -719,17 +433,41 @@ class tile_plugin_t : public wf::plugin_interface_t, wf::per_output_tracker_mixi
         }
     }
 
+    wf::signal::connection_t<view_unmapped_signal> on_view_unmapped = [=] (wf::view_unmapped_signal *ev)
+    {
+        auto toplevel = toplevel_cast(ev->view);
+        if (!toplevel)
+        {
+            return;
+        }
+
+        if (wf::tile::view_node_t::get_node(ev->view))
+        {
+            wf::dassert(toplevel->get_wset() != nullptr);
+
+            auto wo = toplevel->get_output();
+            if (wo && (toplevel->get_wset() == wo->wset()))
+            {
+                wo->get_data<tile_output_plugin_t>()->detach_view(toplevel);
+            } else
+            {
+                tile_workspace_set_data_t::get(toplevel->get_wset()).detach_views(
+                    {wf::tile::view_node_t::get_node(ev->view)});
+            }
+        }
+    };
+
     wf::signal::connection_t<view_pre_moved_to_wset_signal> on_view_pre_moved_to_wset =
         [=] (view_pre_moved_to_wset_signal *ev)
     {
         auto node = wf::tile::view_node_t::get_node(ev->view);
-        if (node)
+        if (node && !preview_manager->is_dragging(ev->view))
         {
             ev->view->store_data(std::make_unique<wf::view_auto_tile_t>());
             if (ev->old_wset)
             {
                 stop_controller(ev->old_wset);
-                tile_workspace_set_data_t::get(ev->old_wset).detach_view(node);
+                tile_workspace_set_data_t::get(ev->old_wset).detach_views({node});
             }
         }
     };
@@ -746,12 +484,12 @@ class tile_plugin_t : public wf::plugin_interface_t, wf::per_output_tracker_mixi
         }
     };
 
-
     wf::signal::connection_t<view_moved_to_wset_signal> on_view_moved_to_wset =
         [=] (view_moved_to_wset_signal *ev)
     {
         if (ev->view->has_data<view_auto_tile_t>() && ev->new_wset)
         {
+            ev->view->erase_data<view_auto_tile_t>();
             stop_controller(ev->new_wset);
             tile_workspace_set_data_t::get(ev->new_wset).attach_view(ev->view);
         }
@@ -766,7 +504,22 @@ class tile_plugin_t : public wf::plugin_interface_t, wf::per_output_tracker_mixi
     {
         output->erase_data<tile_output_plugin_t>();
     }
+
+    ipc::method_callback ipc_get_layout = [=] (const nlohmann::json& params)
+    {
+        return tile::handle_ipc_get_layout(params);
+    };
+
+    ipc::method_callback ipc_set_layout = [=] (nlohmann::json params) -> nlohmann::json
+    {
+        return tile::handle_ipc_set_layout(params);
+    };
 };
+
+std::unique_ptr<tile::tree_node_t>& tile::get_root(wf::workspace_set_t *set, wf::point_t workspace)
+{
+    return tile_workspace_set_data_t::get(set->shared_from_this()).roots[workspace.x][workspace.y];
+}
 }
 
 DECLARE_WAYFIRE_PLUGIN(wf::tile_plugin_t);
