@@ -30,7 +30,7 @@
 #include <wayfire/signal-definitions.hpp>
 #include <wayfire/nonstd/wlroots-full.hpp>
 
-#include "view/surface-impl.hpp"
+#include "wayfire/unstable/wlr-surface-controller.hpp"
 #include "wayfire/scene-input.hpp"
 #include "opengl-priv.hpp"
 #include "seat/input-manager.hpp"
@@ -225,6 +225,7 @@ void wf::compositor_core_impl_t::post_init()
     this->emit(&backend_started_ev);
     this->state = compositor_state_t::RUNNING;
     plugin_mgr  = std::make_unique<wf::plugin_manager_t>();
+    this->bindings->reparse_extensions();
 
     // Move pointer to the middle of the leftmost, topmost output
     wf::pointf_t p;
@@ -251,6 +252,19 @@ void wf::compositor_core_impl_t::shutdown()
     }
 }
 
+void wf::compositor_core_impl_t::disconnect_signals()
+{
+    fini_desktop_apis();
+    fini_xdg_decoration_handlers();
+    drm_lease_request.disconnect();
+    input_inhibit_activated.disconnect();
+    input_inhibit_deactivated.disconnect();
+    idle_inhibitor_created.disconnect();
+    vkbd_created.disconnect();
+    vptr_created.disconnect();
+    pointer_constraint_added.disconnect();
+}
+
 void wf::compositor_core_impl_t::fini()
 {
     this->state = compositor_state_t::SHUTDOWN;
@@ -259,6 +273,9 @@ void wf::compositor_core_impl_t::fini()
 
     LOGI("Unloading plugins...");
     plugin_mgr.reset();
+    // Shut down xwayland first, otherwise, wlroots will attempt to restart it when we kill it via
+    // wl_display_destroy_clients().
+    wf::fini_xwayland();
     LOGI("Stopping clients...");
     wl_display_destroy_clients(static_core->display);
     LOGI("Freeing resources...");
@@ -274,6 +291,7 @@ void wf::compositor_core_impl_t::fini()
     output_layout.reset();
     tx_manager.reset();
     OpenGL::fini();
+    disconnect_signals();
     wl_display_destroy(static_core->display);
 }
 
@@ -478,11 +496,25 @@ std::string wf::compositor_core_impl_t::get_xwayland_display()
     return xwayland_get_display();
 }
 
+void wf::start_move_view_to_wset(wayfire_toplevel_view v, std::shared_ptr<wf::workspace_set_t> new_wset)
+{
+    emit_view_pre_moved_to_wset_pre(v, v->get_wset(), new_wset);
+    if (v->get_wset())
+    {
+        v->get_wset()->remove_view(v);
+        wf::scene::remove_child(v->get_root_node());
+    }
+
+    wf::scene::add_front(new_wset->get_node(), v->get_root_node());
+    new_wset->add_view(v);
+}
+
 void wf::move_view_to_output(wayfire_toplevel_view v, wf::output_t *new_output, bool reconfigure)
 {
+    wf::dassert(!v->parent, "Cannot move a dialog to a different output than its parent!");
+
     auto old_output = v->get_output();
     auto old_wset   = v->get_wset();
-    emit_view_pre_moved_to_wset_pre(v, old_wset, new_output->wset());
 
     uint32_t edges;
     bool fullscreen;
@@ -490,6 +522,9 @@ void wf::move_view_to_output(wayfire_toplevel_view v, wf::output_t *new_output,
     wf::geometry_t old_output_g;
     wf::geometry_t new_output_g;
 
+    int delta_x = 0;
+    int delta_y = 0;
+
     if (reconfigure)
     {
         edges = v->pending_tiled_edges();
@@ -501,19 +536,14 @@ void wf::move_view_to_output(wayfire_toplevel_view v, wf::output_t *new_output,
         auto ratio_y = (double)new_output_g.height / old_output_g.height;
         view_g.x *= ratio_x;
         view_g.y *= ratio_y;
+
+        delta_x = view_g.x - v->get_pending_geometry().x;
+        delta_y = view_g.y - v->get_pending_geometry().y;
     }
 
     assert(new_output);
 
-    if (old_output)
-    {
-        old_output->wset()->remove_view(v);
-        wf::scene::remove_child(v->get_root_node());
-    }
-
-    wf::scene::add_front(new_output->wset()->get_node(), v->get_root_node());
-    new_output->wset()->add_view(v);
-
+    start_move_view_to_wset(v, new_output->wset());
     if (new_output == wf::get_core().seat->get_active_output())
     {
         wf::get_core().seat->focus_view(v);
@@ -532,6 +562,15 @@ void wf::move_view_to_output(wayfire_toplevel_view v, wf::output_t *new_output,
             auto new_g = wf::clamp(view_g, new_output->workarea->get_workarea());
             v->set_geometry(new_g);
         }
+
+        for (auto& dialog : v->enumerate_views())
+        {
+            if ((dialog != v) && (delta_x || delta_y))
+            {
+                dialog->move(dialog->get_pending_geometry().x + delta_x,
+                    dialog->get_pending_geometry().y + delta_y);
+            }
+        }
     }
 
     emit_view_moved_to_wset(v, old_wset, new_output->wset());
