@@ -10,6 +10,7 @@
 #include "../output/output-impl.hpp"
 #include <xf86drmMode.h>
 #include <cstring>
+#include <climits>
 #include <unordered_set>
 #include <drm_fourcc.h>
 #include <wayfire/seat.hpp>
@@ -54,10 +55,18 @@ static wl_output_transform get_transform_from_string(std::string transform)
 }
 
 wlr_output_mode *find_matching_mode(wlr_output *output,
-    const wlr_output_mode& reference)
+    const wlr_output_mode& reference, bool exact_match = false)
 {
     wlr_output_mode *mode;
     wlr_output_mode *best = NULL;
+
+    // Pick highest refresh rate by default
+    int target_refresh = reference.refresh;
+    if (target_refresh == 0)
+    {
+        target_refresh = INT_MAX;
+    }
+
     wl_list_for_each(mode, &output->modes, link)
     {
         if ((mode->width == reference.width) && (mode->height == reference.height))
@@ -67,7 +76,20 @@ wlr_output_mode *find_matching_mode(wlr_output *output,
                 return mode;
             }
 
-            if (!best || (best->refresh < mode->refresh))
+            if ((reference.refresh == 0) && mode->preferred)
+            {
+                // If there is a preferred mode and there is no refresh configured, pick preferred mode.
+                return mode;
+            }
+
+            if (exact_match)
+            {
+                continue;
+            }
+
+            const int best_so_far = best ? std::abs(best->refresh - target_refresh) : INT_MAX;
+            const int current     = std::abs(mode->refresh - target_refresh);
+            if (!best || (best_so_far > current))
             {
                 best = mode;
             }
@@ -236,7 +258,7 @@ static const char *get_format_name(uint32_t format)
 struct output_layout_output_t
 {
     wlr_output *handle;
-    output_state_t current_state;
+    output_state_t current_state{};
     bool is_externally_managed = false;
     bool is_nested_compositor  = false;
     bool inhibited = false;
@@ -484,6 +506,7 @@ struct output_layout_output_t
         this->output =
             std::make_unique<wf::output_impl_t>(handle, effective_size);
         auto wo = output.get();
+        priv_render_manager_start_rendering(wo->render.get());
 
         /* Focus the first output, but do not change the focus on subsequently
          * added outputs. We also change the focus if the noop output was
@@ -583,7 +606,7 @@ struct output_layout_output_t
     }
 
     /** Change the output mode */
-    void apply_mode(const wlr_output_mode& mode)
+    void apply_mode(const wlr_output_mode& mode, bool custom_mode)
     {
         if (handle->current_mode)
         {
@@ -602,7 +625,7 @@ struct output_layout_output_t
         }
 
         refresh_custom_modes();
-        auto built_in = find_matching_mode(handle, mode);
+        auto built_in = find_matching_mode(handle, mode, custom_mode);
         if (built_in)
         {
             wlr_output_set_mode(handle, built_in);
@@ -874,7 +897,7 @@ struct output_layout_output_t
         }
 
         set_enabled(!(state.source & OUTPUT_IMAGE_SOURCE_NONE));
-        apply_mode(state.mode);
+        apply_mode(state.mode, state.uses_custom_mode);
 
         if (state.source & OUTPUT_IMAGE_SOURCE_SELF)
         {
@@ -1056,6 +1079,7 @@ class output_layout_t::impl
                 state.mode.width   = head->state.custom_mode.width;
                 state.mode.height  = head->state.custom_mode.height;
                 state.mode.refresh = head->state.custom_mode.refresh;
+                state.uses_custom_mode = true;
             }
 
             state.position  = {head->state.x, head->state.y};
