@@ -1,43 +1,94 @@
 #include <cstring>
+#include <wayfire/bindings-repository.hpp>
 #include <linux/input-event-codes.h>
+#include <wayland-server-protocol.h>
 #include <xkbcommon/xkbcommon.h>
 
 #include <wayfire/util/log.hpp>
 #include "pointer.hpp"
 #include "keyboard.hpp"
 #include "../core-impl.hpp"
-#include "../../output/output-impl.hpp"
-#include "cursor.hpp"
 #include "touch.hpp"
 #include "input-manager.hpp"
-#include "wayfire/compositor-view.hpp"
+#include "input-method-relay.hpp"
 #include "wayfire/signal-definitions.hpp"
 
 void wf::keyboard_t::setup_listeners()
 {
-    on_config_reload.set_callback([&] (signal_data_t*)
+    on_config_reload = [=] (auto)
     {
         reload_input_options();
-    });
-    wf::get_core().connect_signal("reload-config", &on_config_reload);
+    };
+    wf::get_core().connect(&on_config_reload);
 
     on_key.set_callback([&] (void *data)
     {
-        auto ev   = static_cast<wlr_keyboard_key_event*>(data);
-        auto mode = emit_device_event_signal("keyboard_key", ev);
-
+        auto ev    = static_cast<wlr_keyboard_key_event*>(data);
+        auto mode  = emit_device_event_signal(ev, &handle->base);
         auto& seat = wf::get_core_impl().seat;
-        seat->set_keyboard(this);
+        wf::get_core().seat->notify_activity();
 
-        if (!handle_keyboard_key(ev->keycode, ev->state) &&
-            (mode != input_event_processing_mode_t::NO_CLIENT))
+        if (mode == input_event_processing_mode_t::IGNORE)
         {
-            wlr_seat_keyboard_notify_key(seat->seat,
-                ev->time_msec, ev->keycode, ev->state);
+            emit_device_post_event_signal(ev, &handle->base);
+            return;
         }
 
-        wlr_idle_notify_activity(wf::get_core().protocols.idle, seat->seat);
-        emit_device_event_signal("keyboard_key_post", ev);
+        auto is_im_sent = wf::get_core_impl().im_relay->is_im_sent(handle);
+        if (!is_im_sent)
+        {
+            seat->priv->set_keyboard(this);
+        }
+
+        LOGC(IM, "Received key=", ev->keycode, " state=", ev->state, " im=", is_im_sent);
+        if ((is_im_sent || !handle_keyboard_key(ev->keycode, ev->state)) &&
+            (mode == input_event_processing_mode_t::FULL))
+        {
+            if (!is_im_sent &&
+                wf::get_core_impl().im_relay->handle_key(handle, ev->time_msec, ev->keycode, ev->state))
+            {
+                LOGC(IM, "key=", ev->keycode, " state=", ev->state, " swallowed by IM.");
+                return emit_device_post_event_signal(ev, &handle->base);
+            }
+
+            if (ev->state == WL_KEYBOARD_KEY_STATE_PRESSED)
+            {
+                seat->priv->pressed_keys.insert(ev->keycode);
+            }
+
+            if (ev->state == WL_KEYBOARD_KEY_STATE_RELEASED)
+            {
+                if (seat->priv->pressed_keys.count(ev->keycode))
+                {
+                    seat->priv->pressed_keys.erase(seat->priv->pressed_keys.find(ev->keycode));
+                } else
+                {
+                    LOGC(IM, "Key ", ev->keycode, " ignored.");
+                    return emit_device_post_event_signal(ev, &handle->base);
+                }
+            }
+
+            pre_client_input_event_signal<wlr_keyboard_key_event> pre_ev;
+            pre_ev.event  = ev;
+            pre_ev.device = &handle->base;
+            pre_ev.focus_node = seat->priv->keyboard_focus;
+            wf::get_core().emit(&pre_ev);
+
+            if (pre_ev.carried_out)
+            {
+                LOGC(IM, "key=", ev->keycode, " state=", ev->state, " swallowed by a plugin in pre-client.");
+            } else if (seat->priv->keyboard_focus)
+            {
+                LOGC(IM, "key=", ev->keycode, " state=", ev->state, " sent to node.");
+                seat->priv->keyboard_focus->keyboard_interaction()
+                    .handle_keyboard_key(wf::get_core().seat.get(), *ev);
+            }
+        } else
+        {
+            LOGC(IM, "key=", ev->keycode, " state=", ev->state, " swallowed by a plugin.");
+        }
+
+        emit_device_post_event_signal(ev, &handle->base);
     });
 
     on_modifier.set_callback([&] (void *data)
@@ -45,9 +96,22 @@ void wf::keyboard_t::setup_listeners()
         auto kbd  = static_cast<wlr_keyboard*>(data);
         auto seat = wf::get_core().get_current_seat();
 
-        wlr_seat_set_keyboard(seat, kbd);
-        wlr_seat_keyboard_send_modifiers(seat, &kbd->modifiers);
-        wlr_idle_notify_activity(wf::get_core().protocols.idle, seat);
+        mwlr_keyboard_modifiers_event event;
+        event.time_msec = wf::get_current_time();
+        emit_device_event_signal(&event, &handle->base);
+
+        if (!wf::get_core_impl().im_relay->handle_modifier(kbd))
+        {
+            if (!wf::get_core_impl().im_relay->is_im_sent(handle))
+            {
+                wlr_seat_set_keyboard(seat, kbd);
+            }
+
+            wlr_seat_keyboard_send_modifiers(seat, &kbd->modifiers);
+        }
+
+        wf::get_core().seat->notify_activity();
+        emit_device_post_event_signal(&event, &handle->base);
     });
 
     on_key.connect(&handle->events.key);
@@ -175,14 +239,6 @@ static bool check_vt_switch(wlr_session *session, uint32_t key, uint32_t mods)
         return false;
     }
 
-    /* Somebody inhibited the output, most probably a lockscreen */
-    auto output_impl =
-        static_cast<wf::output_impl_t*>(wf::get_core().get_active_output());
-    if (output_impl->is_inhibited())
-    {
-        return false;
-    }
-
     int target_vt = key - KEY_F1 + 1;
     wlr_session_change_vt(session, target_vt);
 
@@ -270,31 +326,19 @@ bool wf::keyboard_t::handle_keyboard_key(uint32_t key, uint32_t state)
     auto& input = wf::get_core_impl().input;
     auto& seat  = wf::get_core_impl().seat;
 
-    auto active_grab = input->active_grab;
-    bool handled_in_plugin = (active_grab != nullptr);
-    if (active_grab && active_grab->callbacks.keyboard.key)
-    {
-        active_grab->callbacks.keyboard.key(key, state);
-    }
-
+    bool handled_in_plugin = false;
     auto mod = mod_from_key(key);
-    if (mod)
-    {
-        handle_keyboard_mod(mod, state);
-    }
-
     input->locked_mods = this->get_locked_mods();
 
     if (state == WLR_KEY_PRESSED)
     {
-        auto session = wlr_backend_get_session(wf::get_core().backend);
-        if (check_vt_switch(session, key, get_modifiers()))
+        if (check_vt_switch(wf::get_core().session, key, get_modifiers()))
         {
             return true;
         }
 
-        bool modifiers_only = !seat->lpointer->has_pressed_buttons() &&
-            (seat->touch->get_state().fingers.empty()) &&
+        bool modifiers_only = !seat->priv->lpointer->has_pressed_buttons() &&
+            (seat->priv->touch->get_state().fingers.empty()) &&
             this->has_only_modifiers();
 
         /* as long as we have pressed only modifiers, we should check for modifier
@@ -308,7 +352,7 @@ bool wf::keyboard_t::handle_keyboard_key(uint32_t key, uint32_t state)
             mod_binding_key = 0;
         }
 
-        handled_in_plugin |= input->get_active_bindings().handle_key(
+        handled_in_plugin |= wf::get_core().bindings->handle_key(
             wf::keybinding_t{get_modifiers(), key}, mod_binding_key);
     } else
     {
@@ -321,7 +365,7 @@ bool wf::keyboard_t::handle_keyboard_key(uint32_t key, uint32_t state)
 
             if ((timeout <= 0) || (time_elapsed < milliseconds(timeout)))
             {
-                handled_in_plugin |= input->get_active_bindings().handle_key(
+                wf::get_core().bindings->handle_key(
                     wf::keybinding_t{get_modifiers() | mod, 0}, mod_binding_key);
             }
         }
@@ -329,20 +373,5 @@ bool wf::keyboard_t::handle_keyboard_key(uint32_t key, uint32_t state)
         mod_binding_key = 0;
     }
 
-    auto iv = interactive_view_from_view(seat->keyboard_focus.get());
-    if (iv && !handled_in_plugin)
-    {
-        iv->handle_key(key, state);
-    }
-
     return handled_in_plugin;
 }
-
-void wf::keyboard_t::handle_keyboard_mod(uint32_t modifier, uint32_t state)
-{
-    auto active_grab = wf::get_core_impl().input->active_grab;
-    if (active_grab && active_grab->callbacks.keyboard.mod)
-    {
-        active_grab->callbacks.keyboard.mod(modifier, state);
-    }
-}
