@@ -1,18 +1,25 @@
+#include <wayfire/bindings-repository.hpp>
 #include "tablet.hpp"
 #include "../core-impl.hpp"
-#include "../wm.hpp"
 #include "pointer.hpp"
 #include "cursor.hpp"
-#include "input-manager.hpp"
+#include "view/view-impl.hpp"
+#include "wayfire/core.hpp"
+#include "wayfire/debug.hpp"
+#include "wayfire/scene-input.hpp"
+#include "wayfire/view.hpp"
+#include <algorithm>
+#include <memory>
 #include <wayfire/signal-definitions.hpp>
 #include <wayfire/output-layout.hpp>
 #include <linux/input-event-codes.h>
+#include <wayfire/window-manager.hpp>
 
 /* --------------------- Tablet tool implementation ------------------------- */
 wf::tablet_tool_t::tablet_tool_t(wlr_tablet_tool *tool,
-    wlr_tablet_v2_tablet *tablet)
+    wlr_tablet_v2_tablet *tablet_v2)
 {
-    this->tablet_v2 = tablet;
+    this->tablet_v2 = tablet_v2;
 
     /* Initialize tool_v2 */
     this->tool = tool;
@@ -26,36 +33,32 @@ wf::tablet_tool_t::tablet_tool_t(wlr_tablet_tool *tool,
     this->on_destroy.set_callback([=] (void*)
     {
         this->tool->data = nullptr;
-        delete this;
+        tablet->tools_list.erase(
+            std::remove_if(tablet->tools_list.begin(), tablet->tools_list.end(),
+                [this] (const auto& p) { return p.get() == this; }),
+            tablet->tools_list.end());
     });
     this->on_destroy.connect(&tool->events.destroy);
 
     /* Ungrab surface, and update focused surface if a surface is unmapped,
      * we don't want to end up with a reference to unfocused or a destroyed
      * surface. */
-    on_surface_map_state_changed.set_callback([=] (signal_data_t *data)
+    on_root_node_updated = [=] (wf::scene::root_node_update_signal *data)
     {
-        auto ev = static_cast<surface_map_state_changed_signal*>(data);
-        if (!ev->surface->is_mapped() && (ev->surface == this->grabbed_surface))
+        if (!(data->flags & wf::scene::update_flag::INPUT_STATE))
         {
-            this->grabbed_surface = nullptr;
+            return;
         }
 
-        update_tool_position();
-    });
+        if (grabbed_node && !is_grabbed_node_alive(grabbed_node))
+        {
+            reset_grab();
+        }
 
-    wf::get_core().connect_signal("surface-mapped",
-        &on_surface_map_state_changed);
-    wf::get_core().connect_signal("surface-unmapped",
-        &on_surface_map_state_changed);
+        update_tool_position(false);
+    };
 
-    on_views_updated.set_callback([&] (wf::signal_data_t *data)
-    {
-        update_tool_position();
-    });
-
-    wf::get_core().connect_signal("output-stack-order-changed", &on_views_updated);
-    wf::get_core().connect_signal("view-geometry-changed", &on_views_updated);
+    wf::get_core().scene()->connect(&on_root_node_updated);
 
     /* Just pass cursor set requests to core, but translate them to
      * regular pointer set requests */
@@ -87,7 +90,7 @@ wf::tablet_tool_t::tablet_tool_t(wlr_tablet_tool *tool,
         pev.hotspot_y = ev->hotspot_y;
         pev.serial    = ev->serial;
         pev.seat_client = ev->seat_client;
-        wf::get_core_impl().seat->cursor->set_cursor(&pev, false);
+        wf::get_core_impl().seat->priv->cursor->set_cursor(&pev, false);
     });
     on_set_cursor.connect(&tool_v2->events.set_cursor);
 }
@@ -97,48 +100,57 @@ wf::tablet_tool_t::~tablet_tool_t()
     tool->data = NULL;
 }
 
-void wf::tablet_tool_t::update_tool_position()
+static inline wlr_surface *wlr_surface_from_node(wf::scene::node_ptr node)
+{
+    if (auto snode = dynamic_cast<wf::scene::wlr_surface_node_t*>(node.get()))
+    {
+        return snode->get_surface();
+    }
+
+    return nullptr;
+}
+
+void wf::tablet_tool_t::update_tool_position(bool real_update)
 {
     if (!is_active)
     {
         return;
     }
 
-    auto& core  = wf::get_core_impl();
-    auto& input = core.input;
-    auto gc     = core.get_cursor_position();
-
-    /* XXX: tablet input works only with programs, Wayfire itself doesn't do
-     * anything useful with it */
-    if (core.input->input_grabbed())
-    {
-        return;
-    }
+    auto& core = wf::get_core_impl();
+    auto gc    = core.get_cursor_position();
 
     /* Figure out what surface is under the tool */
     wf::pointf_t local; // local to the surface
-    wf::surface_interface_t *surface = nullptr;
-    if (this->grabbed_surface)
+    wf::scene::node_ptr focus_node = nullptr;
+    if (this->grabbed_node)
     {
-        surface = this->grabbed_surface;
-        local   = get_surface_relative_coords(surface, gc);
+        focus_node = this->grabbed_node;
+        local = get_node_local_coords(focus_node.get(), gc);
     } else
     {
-        surface = input->input_surface_at(gc, local);
+        auto input_node = wf::get_core().scene()->find_node_at(gc);
+        if (input_node)
+        {
+            focus_node = input_node->node->shared_from_this();
+            local = input_node->local_coords;
+        }
     }
 
-    set_focus(surface);
+    bool focus_changed = set_focus(focus_node);
 
     /* If focus is a wlr surface, send position */
-    wlr_surface *next_focus = surface ? surface->get_wlr_surface() : nullptr;
-    if (next_focus)
+    wlr_surface *next_focus = wlr_surface_from_node(focus_node);
+    if (next_focus && (real_update || focus_changed))
     {
         wlr_tablet_v2_tablet_tool_notify_motion(tool_v2, local.x, local.y);
     }
 }
 
-void wf::tablet_tool_t::set_focus(wf::surface_interface_t *surface)
+bool wf::tablet_tool_t::set_focus(wf::scene::node_ptr surface)
 {
+    bool focus_changed = surface != this->proximity_surface;
+
     /* Unfocus old surface */
     if ((surface != this->proximity_surface) && this->proximity_surface)
     {
@@ -147,19 +159,30 @@ void wf::tablet_tool_t::set_focus(wf::surface_interface_t *surface)
     }
 
     /* Set the new focus, if it is a wlr surface */
-    wf::get_core_impl().seat->ensure_input_surface(surface);
-    wlr_surface *next_focus = surface ? surface->get_wlr_surface() : nullptr;
+
+    wlr_surface *next_focus = wlr_surface_from_node(surface);
+    if (next_focus)
+    {
+        wf::xwayland_bring_to_front(next_focus);
+    }
+
     if (next_focus && wlr_surface_accepts_tablet_v2(tablet_v2, next_focus))
     {
         this->proximity_surface = surface;
-        wlr_tablet_v2_tablet_tool_notify_proximity_in(
-            tool_v2, tablet_v2, next_focus);
+        wlr_tablet_v2_tablet_tool_notify_proximity_in(tool_v2, tablet_v2, next_focus);
     }
 
     if (!next_focus)
     {
         wf::get_core().set_cursor("default");
     }
+
+    return focus_changed;
+}
+
+void wf::tablet_tool_t::reset_grab()
+{
+    this->grabbed_node = nullptr;
 }
 
 void wf::tablet_tool_t::passthrough_axis(wlr_tablet_tool_axis_event *ev)
@@ -218,21 +241,16 @@ void wf::tablet_tool_t::handle_tip(wlr_tablet_tool_tip_event *ev)
     if (ev->state == WLR_TABLET_TOOL_TIP_DOWN)
     {
         wlr_send_tablet_v2_tablet_tool_down(tool_v2);
-        this->grabbed_surface = this->proximity_surface;
+        this->grabbed_node = this->proximity_surface;
 
         /* Try to focus the view under the tool */
-        auto view = dynamic_cast<wf::view_interface_t*>(
-            this->proximity_surface->get_main_surface());
-        if (view)
-        {
-            wm_focus_request data;
-            data.surface = this->proximity_surface;
-            view->get_output()->emit_signal("wm-focus-request", &data);
-        }
+        auto view = wf::node_to_view(this->proximity_surface);
+        wf::get_core().default_wm->focus_raise_view(view);
     } else
     {
         wlr_send_tablet_v2_tablet_tool_up(tool_v2);
-        this->grabbed_surface = nullptr;
+        this->grabbed_node = nullptr;
+        update_tool_position(false);
     }
 }
 
@@ -252,7 +270,7 @@ void wf::tablet_tool_t::handle_proximity(wlr_tablet_tool_proximity_event *ev)
     } else
     {
         is_active = true;
-        update_tool_position();
+        update_tool_position(true);
     }
 }
 
@@ -278,7 +296,11 @@ wf::tablet_tool_t*wf::tablet_t::ensure_tool(wlr_tablet_tool *tool)
 {
     if (tool->data == NULL)
     {
-        new wf::tablet_tool_t(tool, tablet_v2);
+        auto wtool = std::make_unique<wf::tablet_tool_t>(tool, tablet_v2);
+        wtool->tablet = this;
+        auto ptr = wtool.get();
+        this->tools_list.push_back(std::move(wtool));
+        return ptr;
     }
 
     return (wf::tablet_tool_t*)tool->data;
@@ -287,33 +309,24 @@ wf::tablet_tool_t*wf::tablet_t::ensure_tool(wlr_tablet_tool *tool)
 void wf::tablet_t::handle_tip(wlr_tablet_tool_tip_event *ev,
     input_event_processing_mode_t mode)
 {
-    auto& input = wf::get_core_impl().input;
-    auto& seat  = wf::get_core_impl().seat;
-    seat->break_mod_bindings();
+    if (should_use_absolute_positioning(ev->tool))
+    {
+        wlr_cursor_warp_absolute(cursor, &ev->tablet->base, ev->x, ev->y);
+    }
+
+    auto& seat = wf::get_core_impl().seat;
+    seat->priv->break_mod_bindings();
 
     bool handled_in_binding = false;
     if (ev->state == WLR_TABLET_TOOL_TIP_DOWN)
     {
-        auto gc     = seat->cursor->get_cursor_position();
+        auto gc     = seat->priv->cursor->get_cursor_position();
         auto output =
             wf::get_core().output_layout->get_output_at(gc.x, gc.y);
-        wf::get_core().focus_output(output);
+        wf::get_core().seat->focus_output(output);
 
-        handled_in_binding |= input->get_active_bindings().handle_button(
-            wf::buttonbinding_t{seat->get_modifiers(), BTN_LEFT});
-    }
-
-    if (input->input_grabbed())
-    {
-        /* Simulate buttons, in case some application started moving */
-        if (input->active_grab->callbacks.pointer.button)
-        {
-            uint32_t state = ev->state == WLR_TABLET_TOOL_TIP_DOWN ?
-                WLR_BUTTON_PRESSED : WLR_BUTTON_RELEASED;
-            input->active_grab->callbacks.pointer.button(BTN_LEFT, state);
-        }
-
-        return;
+        handled_in_binding |= wf::get_core().bindings->handle_button(
+            wf::buttonbinding_t{seat->priv->get_modifiers(), BTN_LEFT});
     }
 
     auto tool = ensure_tool(ev->tool);
@@ -326,36 +339,20 @@ void wf::tablet_t::handle_tip(wlr_tablet_tool_tip_event *ev,
 void wf::tablet_t::handle_axis(wlr_tablet_tool_axis_event *ev,
     input_event_processing_mode_t mode)
 {
-    auto& input = wf::get_core_impl().input;
-
     /* Update cursor position */
-    switch (ev->tool->type)
+    if (should_use_absolute_positioning(ev->tool))
     {
-      case WLR_TABLET_TOOL_TYPE_MOUSE:
-        wlr_cursor_move(cursor, &ev->tablet->base, ev->dx, ev->dy);
-        break;
-
-      default:
         double x = (ev->updated_axes & WLR_TABLET_TOOL_AXIS_X) ? ev->x : NAN;
         double y = (ev->updated_axes & WLR_TABLET_TOOL_AXIS_Y) ? ev->y : NAN;
         wlr_cursor_warp_absolute(cursor, &ev->tablet->base, x, y);
-    }
-
-    if (input->input_grabbed())
+    } else
     {
-        /* Simulate movement */
-        if (input->active_grab->callbacks.pointer.motion)
-        {
-            auto gc = wf::get_core().get_cursor_position();
-            input->active_grab->callbacks.pointer.motion(gc.x, gc.y);
-        }
-
-        return;
+        wlr_cursor_move(cursor, &ev->tablet->base, ev->dx, ev->dy);
     }
 
     /* Update focus */
     auto tool = ensure_tool(ev->tool);
-    tool->update_tool_position();
+    tool->update_tool_position(true);
     tool->passthrough_axis(ev);
 }
 
@@ -369,19 +366,23 @@ void wf::tablet_t::handle_button(wlr_tablet_tool_button_event *ev,
 void wf::tablet_t::handle_proximity(wlr_tablet_tool_proximity_event *ev,
     input_event_processing_mode_t mode)
 {
-    ensure_tool(ev->tool)->handle_proximity(ev);
+    if (should_use_absolute_positioning(ev->tool))
+    {
+        wlr_cursor_warp_absolute(cursor, &ev->tablet->base, ev->x, ev->y);
+    }
 
+    ensure_tool(ev->tool)->handle_proximity(ev);
     auto& impl = wf::get_core_impl();
 
     /* Show appropriate cursor */
     if (ev->state == WLR_TABLET_TOOL_PROXIMITY_OUT)
     {
         impl.set_cursor("default");
-        impl.seat->lpointer->set_enable_focus(true);
+        impl.seat->priv->lpointer->set_enable_focus(true);
     } else
     {
         wf::get_core().set_cursor("crosshair");
-        impl.seat->lpointer->set_enable_focus(false);
+        impl.seat->priv->lpointer->set_enable_focus(false);
     }
 }
 
@@ -393,30 +394,35 @@ wf::tablet_pad_t::tablet_pad_t(wlr_input_device *pad) :
     this->pad_v2 = wlr_tablet_pad_create(core.protocols.tablet_v2,
         core.get_current_seat(), pad);
 
-    on_input_devices_changed = [=] (void*)
+
+    on_device_added = [=] (auto)
     {
         select_default_tool();
     };
 
-    wf::get_core().connect_signal("input-device-added",
-        &on_input_devices_changed);
-    wf::get_core().connect_signal("input-device-removed",
-        &on_input_devices_changed);
-
-    on_keyboard_focus_changed = [=] (void*)
+    on_device_removed = [=] (auto)
     {
-        update_focus();
+        select_default_tool();
     };
 
-    wf::get_core().connect_signal("keyboard-focus-changed",
-        &on_keyboard_focus_changed);
+    wf::get_core().connect(&on_device_added);
+    wf::get_core().connect(&on_device_removed);
 
+    on_keyboard_focus_changed.set_callback([=] (auto)
+    {
+        update_focus();
+    });
+    wf::get_core().connect(&on_keyboard_focus_changed);
     select_default_tool();
 
     on_attach.set_callback([=] (void *data)
     {
-        auto tablet = static_cast<wlr_tablet_tool*>(data);
-        attach_to_tablet((tablet_t*)tablet->data);
+        auto wlr_tool = static_cast<wlr_tablet_tool*>(data);
+        auto tool     = (tablet_tool_t*)wlr_tool->data;
+        if (tool)
+        {
+            attach_to_tablet(tool->tablet);
+        }
     });
 
     on_button.set_callback([=] (void *data)
@@ -447,24 +453,14 @@ wf::tablet_pad_t::tablet_pad_t(wlr_input_device *pad) :
     on_button.connect(&wlr_tablet_pad_from_input_device(pad)->events.button);
     on_strip.connect(&wlr_tablet_pad_from_input_device(pad)->events.strip);
     on_ring.connect(&wlr_tablet_pad_from_input_device(pad)->events.ring);
-}
 
-wf::tablet_pad_t::~tablet_pad_t()
-{
-    wf::get_core().disconnect_signal("input-device-added",
-        &on_input_devices_changed);
-    wf::get_core().disconnect_signal("input-device-removed",
-        &on_input_devices_changed);
-    wf::get_core().disconnect_signal("keyboard-focus-changed",
-        &on_keyboard_focus_changed);
+    on_focus_destroy.set_callback([&] (auto) { update_focus(nullptr); });
 }
 
 void wf::tablet_pad_t::update_focus()
 {
-    auto active_output = wf::get_core().get_active_output();
-    auto focus_view    =
-        active_output ? active_output->get_active_view() : nullptr;
-    auto focus_surface = focus_view ? focus_view->priv->wsurface : nullptr;
+    auto focus_view    = wf::get_core().seat->get_active_view();
+    auto focus_surface = focus_view ? focus_view->get_wlr_surface() : nullptr;
     update_focus(focus_surface);
 }
 
@@ -475,15 +471,22 @@ void wf::tablet_pad_t::update_focus(wlr_surface *focus_surface)
         return;
     }
 
-    if (focus_surface && attached_to)
-    {
-        wlr_tablet_v2_tablet_pad_notify_enter(pad_v2,
-            attached_to->tablet_v2, focus_surface);
-    } else
+    if (old_focus)
     {
         wlr_tablet_v2_tablet_pad_notify_leave(pad_v2, old_focus);
     }
 
+    if (focus_surface && attached_to)
+    {
+        wlr_tablet_v2_tablet_pad_notify_enter(pad_v2, attached_to->tablet_v2, focus_surface);
+    }
+
+    on_focus_destroy.disconnect();
+    if (focus_surface)
+    {
+        on_focus_destroy.connect(&focus_surface->events.destroy);
+    }
+
     old_focus = focus_surface;
 }
 
@@ -524,7 +527,7 @@ void wf::tablet_pad_t::select_default_tool()
         auto pad_gr = get_group(this->get_wlr_handle());
         auto tab_gr = get_group(dev->get_wlr_handle());
 
-        if (pad_gr && tab_gr && (pad_gr == tab_gr))
+        if (pad_gr == tab_gr)
         {
             attach_to_tablet(static_cast<tablet_t*>(dev.get()));
             return;
@@ -533,3 +536,20 @@ void wf::tablet_pad_t::select_default_tool()
 
     attach_to_tablet(nullptr);
 }
+
+bool wf::tablet_t::should_use_absolute_positioning(wlr_tablet_tool *tool)
+{
+    static wf::option_wrapper_t<std::string> tablet_motion_mode{"input/tablet_motion_mode"};
+
+    /* Update cursor position */
+    if ((std::string)tablet_motion_mode == "absolute")
+    {
+        return true;
+    } else if ((std::string)tablet_motion_mode == "relative")
+    {
+        return false;
+    } else
+    {
+        return tool->type != WLR_TABLET_TOOL_TYPE_MOUSE;
+    }
+}
