@@ -106,6 +106,7 @@ static void wlr_log_handler(wlr_log_importance level,
     wf::log::log_plain(wlevel, buffer);
 }
 
+static std::optional<int> exit_because_signal;
 static void signal_handler(int signal)
 {
     std::string error;
@@ -124,12 +125,12 @@ static void signal_handler(int signal)
         break;
 
       case SIGINT:
-        LOGI("Got SIGINT, shutting down");
+        exit_because_signal = SIGINT;
         wf::get_core().shutdown();
         return;
 
       case SIGTERM:
-        LOGI("Got SIGTERM, shutting down");
+        exit_because_signal = SIGTERM;
         wf::get_core().shutdown();
         return;
 
@@ -227,6 +228,10 @@ void parse_extended_debugging(const std::vector<std::string>& categories)
         {
             LOGD("Enabling extended debugging for input method events");
             wf::log::enabled_categories.set((size_t)wf::log::logging_category::IM, 1);
+        } else if (cat == "render")
+        {
+            LOGD("Enabling extended debugging for render events");
+            wf::log::enabled_categories.set((size_t)wf::log::logging_category::RENDER, 1);
         } else
         {
             LOGE("Unrecognized debugging category \"", cat, "\"");
@@ -234,6 +239,17 @@ void parse_extended_debugging(const std::vector<std::string>& categories)
     }
 }
 
+// Override assert() handler to be more useful and print a trace.
+// extern "C" {
+// void __assert_fail(const char* expr, const char *filename, unsigned int line, const char *assert_func)
+// {
+// LOGE("Assertion failed: ", expr, " at ", filename, ":", line, " in ", assert_func);
+// wf::print_trace(false);
+// std::_Exit(-1);
+// }
+// }
+
+// Override exception handling to be more useful and print a trace.
 // extern "C" {
 // void __cxa_throw(void *ex, void *info, void (*dest)(void*))
 // {
@@ -430,6 +446,14 @@ int main(int argc, char *argv[])
     core.post_init();
 
     wl_display_run(core.display);
+    if (exit_because_signal == SIGINT)
+    {
+        LOGI("Got SIGINT, shutting down");
+    } else if (exit_because_signal == SIGTERM)
+    {
+        LOGI("Got SIGTERM, shutting down");
+    }
+
     wf::compositor_core_impl_t::deallocate_core();
     LOGI("Shutdown successful!");
     return EXIT_SUCCESS;
