@@ -17,7 +17,6 @@
 #include <wayfire/util/log.hpp>
 #include <wayfire/nonstd/wlroots-full.hpp>
 #include "wayfire/output.hpp"
-#include "wayfire/workspace-set.hpp"
 #include "wayfire/output-layout.hpp"
 #include "../view-impl.hpp"
 #include "wlr-layer-shell-unstable-v1-protocol.h"
@@ -224,18 +223,6 @@ struct wf_layer_shell_manager
         return result;
     }
 
-    layer_t filter_views(wf::output_t *output)
-    {
-        layer_t result;
-        for (int i = 0; i < 4; i++)
-        {
-            auto layer_result = filter_views(output, i);
-            result.insert(result.end(), layer_result.begin(), layer_result.end());
-        }
-
-        return result;
-    }
-
     void set_exclusive_zone(wayfire_layer_shell_view *v)
     {
         int edges = v->lsurface->current.anchor;
@@ -327,10 +314,9 @@ struct wf_layer_shell_manager
         v->configure(box);
     }
 
-    void arrange_layer(wf::output_t *output, int layer)
+    void arrange_exclusive_zone(wf::output_t *output, int layer)
     {
         auto views = filter_views(output, layer);
-
         /* First we need to put all views that have exclusive zone set.
          * The rest are then placed into the free area */
         for (auto v : views)
@@ -345,7 +331,11 @@ struct wf_layer_shell_manager
                 v->remove_anchored(false);
             }
         }
+    }
 
+    void arrange_floating(wf::output_t *output, int layer)
+    {
+        auto views = filter_views(output, layer);
         auto usable_workarea = output->workarea->get_workarea();
         for (auto v : views)
         {
@@ -371,13 +361,24 @@ struct wf_layer_shell_manager
 
     void arrange_layers(wf::output_t *output)
     {
-        auto views = filter_views(output);
+        const auto layers = {
+            ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY,
+            ZWLR_LAYER_SHELL_V1_LAYER_TOP,
+            ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM,
+            ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND
+        };
+
+        for (auto& layer : layers)
+        {
+            arrange_exclusive_zone(output, layer);
+        }
 
-        arrange_layer(output, ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY);
-        arrange_layer(output, ZWLR_LAYER_SHELL_V1_LAYER_TOP);
-        arrange_layer(output, ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM);
-        arrange_layer(output, ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND);
         output->workarea->reflow_reserved_areas();
+
+        for (auto& layer : layers)
+        {
+            arrange_floating(output, layer);
+        }
     }
 };
 
@@ -499,7 +500,8 @@ void wayfire_layer_shell_view::map()
     on_commit_unmapped.disconnect();
 
     priv->set_mapped_surface_contents(main_surface);
-    priv->set_mapped(true);
+    priv->set_mapped(main_surface->get_surface());
+    priv->set_enabled(true);
     on_surface_commit.connect(&lsurface->surface->events.commit);
 
     /* Read initial data */
@@ -523,10 +525,10 @@ void wayfire_layer_shell_view::unmap()
 
     emit_view_pre_unmap();
     priv->unset_mapped_surface_contents();
+    priv->set_mapped(nullptr);
     on_surface_commit.disconnect();
     emit_view_unmap();
-    priv->set_mapped(false);
-
+    priv->set_enabled(false);
     wf_layer_shell_manager::get_instance().handle_unmap(this);
 }
 
@@ -673,19 +675,24 @@ class layer_shell_view_controller_t
 };
 
 static wlr_layer_shell_v1 *layer_shell_handle;
+static wf::wl_listener_wrapper on_layer_shell_surface_created;
+
 void wf::init_layer_shell()
 {
-    static wf::wl_listener_wrapper on_created;
-
     layer_shell_handle = wlr_layer_shell_v1_create(wf::get_core().display, 4);
     if (layer_shell_handle)
     {
-        on_created.set_callback([] (void *data)
+        on_layer_shell_surface_created.set_callback([] (void *data)
         {
             auto lsurf = static_cast<wlr_layer_surface_v1*>(data);
             new layer_shell_view_controller_t{lsurf};
         });
 
-        on_created.connect(&layer_shell_handle->events.new_surface);
+        on_layer_shell_surface_created.connect(&layer_shell_handle->events.new_surface);
     }
 }
+
+void wf::fini_layer_shell()
+{
+    on_layer_shell_surface_created.disconnect();
+}
