@@ -7,14 +7,12 @@
 #include <wayfire/opengl.hpp>
 #include <wayfire/toplevel.hpp>
 
-#include "surface-impl.hpp"
 #include "wayfire/core.hpp"
 #include "wayfire/nonstd/tracking-allocator.hpp"
 #include "wayfire/signal-provider.hpp"
 #include "wayfire/unstable/wlr-surface-node.hpp"
 #include "wayfire/output.hpp"
 #include "wayfire/scene.hpp"
-#include "wayfire/util.hpp"
 #include "wayfire/view-transform.hpp"
 #include <wayfire/nonstd/wlroots-full.hpp>
 #include <wayfire/compositor-view.hpp>
@@ -46,7 +44,8 @@ class view_interface_t::view_priv_impl
     scene::floating_inner_ptr surface_root_node;
     wf::output_t *output;
 
-    void set_mapped(bool mapped);
+    void set_mapped(wlr_surface *surface);
+    void set_enabled(bool enabled);
     void set_mapped_surface_contents(std::shared_ptr<scene::wlr_surface_node_t> content);
     void unset_mapped_surface_contents();
 
@@ -60,10 +59,17 @@ class view_interface_t::view_priv_impl
  */
 void adjust_geometry_for_gravity(wf::toplevel_state_t& desired_state, wf::dimensions_t actual_size);
 
+void adjust_view_output_on_map(wf::toplevel_view_interface_t *self);
+void adjust_view_pending_geometry_on_start_map(wf::toplevel_view_interface_t *self,
+    wf::geometry_t map_geometry_client, bool map_fs, bool map_maximized);
+
 /** Emit the map signal for the given view */
 void init_xdg_shell();
-void init_xwayland();
+void init_xwayland(bool lazy);
 void init_layer_shell();
+void fini_xdg_shell();
+void fini_xwayland();
+void fini_layer_shell();
 
 std::string xwayland_get_display();
 void xwayland_update_default_cursor();
@@ -73,7 +79,9 @@ void xwayland_bring_to_front(wlr_surface *surface);
 int xwayland_get_pid();
 
 void init_desktop_apis();
+void fini_desktop_apis();
 void init_xdg_decoration_handlers();
+void fini_xdg_decoration_handlers();
 
 pointf_t place_popup_at(wlr_surface *parent, wlr_surface *popup, wf::pointf_t relative);
 }
